import threading
import traceback
from time import time

from ..domain import ThreadTask
from ........domain.enums.events import EVENT_LOG
from ........pubsub.base import ChannelQueue
from ........pubsub.domain import TaskMessage
from ........pubsub.publisher import Publisher


class ParallelWorker(threading.Thread):
    def __init__(
            self,
            thread_id: int,
            operation_integration,
            channel: ChannelQueue,
            process_channel: ChannelQueue,
            process_result_channel: ChannelQueue,
            target_method,
            target_kwargs,
            *args,
            **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.operation_integration = operation_integration
        self.publisher = Publisher(channel=channel)
        self.target_kwargs = target_kwargs
        self.target_method = target_method
        self.process_result_channel = process_result_channel
        self.process_channel = process_channel
        self.thread_id = thread_id
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

    def run(self):

        self.publisher.publish(
            message=TaskMessage(
                event=EVENT_LOG,
                kwargs={
                    'data': self.operation_integration,
                    'message': f"{self.thread_id} process started"
                }
            )
        )
        try:
            while True:
                process_task: ThreadTask = self.process_channel.get()
                process_task.Id = self.thread_id
                process_task.State = 2
                if process_task.IsFinished:
                    self.publisher.publish(
                        message=TaskMessage(
                            event=EVENT_LOG,
                            kwargs={
                                'data': self.operation_integration,
                                'message': f"{self.thread_id} process finished"
                            }
                        )
                    )
                    break
                else:
                    start = time()
                    # calling target method
                    self.target_kwargs["thread_id"] = self.thread_id
                    result = self.target_method(**self.target_kwargs)
                    end = time()

                    self.publisher.publish(
                        message=TaskMessage(
                            event=EVENT_LOG,
                            kwargs={
                                'data': self.operation_integration,
                                'message': f"{self.thread_id} process finished. time:{end - start}"
                            }
                        )
                    )
                    process_task.IsFinished = True
                    process_task.State = 3
                    process_task.Result = result
                    self.process_result_channel.put(process_task)
                    return
        except Exception as ex:

            self.publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': self.operation_integration,
                        'message': f"{self.thread_id} process getting error.",
                        'exception': ex
                    }
                )
            )
            process_task = ThreadTask(
                Id=self.thread_id,
                State=4,
                Exception=ex,
                Traceback=traceback.format_exc(),
                IsFinished=True
            )
            self.process_result_channel.put(process_task)



import json
import logging
from json import dumps
from queue import Queue
from typing import List

from kafka import KafkaProducer, KafkaAdminClient, KafkaConsumer
from kafka.admin import NewTopic
from kafka.errors import TopicAlreadyExistsError
from pandas import DataFrame

from pdip.integrator.connection.domain import DataQueueTask
from ..base import QueueConnector

kafka_logger = logging.getLogger(name="kafka")
kafka_logger.setLevel(level=logging.WARNING)


class KafkaConnector(QueueConnector):
    def __init__(self,
                 servers: List[str],
                 auth):
        self.auth = auth
        self.servers = servers
        self.data_frame = None
        self.client_id = "pdi_kafka_client"
        self.__admin: KafkaAdminClient = None
        self.__producer: KafkaProducer = None
        self.__consumer: KafkaConsumer = None
        self.connect()

    def connect(self):
        # KAFKA_CLIENT_ID: str = "client_id"
        self.create_admin_client()
        self.create_producer()
        # self.__admin = KafkaAdminClient(bootstrap_servers=self.servers,
        #                                 **self.auth
        #                                 )
        # os.path.isdir(self.folder)
        pass

    def disconnect(self):
        pass
        # try:
        #     if self.file is not None:
        #         self.file.close()
        # except Exception:
        #     pass

    def create_admin_client(self):
        if self.auth is not None:
            self.__admin = KafkaAdminClient(bootstrap_servers=self.servers,
                                            **self.auth
                                            )
        elif self.auth is None:
            self.__admin = KafkaAdminClient(bootstrap_servers=self.servers,
                                            )

    def create_producer(self):
        if self.auth is not None:
            self.__producer = KafkaProducer(bootstrap_servers=list(self.servers),
                                            client_id=self.client_id,
                                            value_serializer=lambda x: dumps(x).encode('utf-8'),
                                            **self.auth)
        elif self.auth is None:
            self.__producer = KafkaProducer(bootstrap_servers=list(self.servers),
                                            client_id=self.client_id,
                                            value_serializer=lambda x: dumps(x).encode('utf-8'))

    def create_consumer(self, topic_name: str, auto_offset_reset: str, enable_auto_commit: bool, group_id: str):
        # KAFKA_CONSUMER_GROUP_ID: str = "consumer_group_id"
        # KAFKA_AUTO_OFFSET_RESET: str = "earliest"
        # KAFKA_ENABLE_AUTO_COMMIT: bool = True
        if self.auth is not None:
            self.__consumer = KafkaConsumer(topic_name, bootstrap_servers=list(self.servers),
                                            consumer_timeout_ms=5000,
                                            auto_offset_reset=auto_offset_reset,
                                            enable_auto_commit=enable_auto_commit,
                                            group_id=group_id,
                                            value_deserializer=lambda x: json.loads(x.decode('utf-8')),
                                            **self.auth)
        if self.auth is None:
            self.__consumer = KafkaConsumer(topic_name, bootstrap_servers=list(self.servers),
                                            consumer_timeout_ms=5000,
                                            auto_offset_reset=auto_offset_reset,
                                            enable_auto_commit=enable_auto_commit,
                                            group_id=group_id,
                                            value_deserializer=lambda x: json.loads(x.decode('utf-8')))

    def get_unpredicted_data(self, limit: int, process_count: int, data_queue: Queue, result_queue: Queue) -> DataFrame:
        data = []
        total_data_count = 0
        limited_data_count = 0
        transmitted_data_count = 0
        task_id = 0
        for message in self.__consumer:
            data.append(message.value)
            total_data_count = total_data_count + 1
            limited_data_count = limited_data_count + 1
            if limited_data_count >= limit:
                task_id = task_id + 1
                data_queue_task = DataQueueTask(Id=task_id, Data=data, Start=total_data_count - limited_data_count,
                                                End=total_data_count, Limit=limit, IsFinished=False)
                data_queue.put(data_queue_task)
                transmitted_data_count = transmitted_data_count + 1
                limited_data_count = 0
                data = []
                if transmitted_data_count >= process_count:
                    result = result_queue.get()
                    if result:
                        transmitted_data_count = transmitted_data_count - 1
                    else:
                        break

        if data is not None and len(data) > 0:
            task_id = task_id + 1
            total_data_count = total_data_count + len(data)
            data_queue_task = DataQueueTask(Id=task_id, Data=data, Start=total_data_count - limited_data_count,
                                            End=total_data_count, Limit=limit, IsFinished=False)
            data_queue.put(data_queue_task)

    def get_data(self, limit: int) -> DataFrame:
        data = []
        data_count = 0
        for message in self.__consumer:

            data.append(message.value)
            data_count = data_count + 1
            if data_count >= limit:
                break
        df = DataFrame(data)
        return df

    def create_topic(self, topic_name):
        # KAFKA_TOPIC_NUM_PARTITIONS: int = 3
        # KAFKA_TOPIC_REPLICA_FACTOR: int = 3
        if not self.topic_exists(topic_name=topic_name):
            try:
                topic = [NewTopic(name=topic_name, num_partitions=1, replication_factor=1)]
                response = self.__admin.create_topics(new_topics=topic, validate_only=False)
            except TopicAlreadyExistsError as ex:
                print(f"{topic_name} Topic Exist")

    def topic_exists(self, topic_name) -> bool:
        topic_metadata = self.__admin.list_topics()
        return topic_name in topic_metadata

    def delete_topic(self, topic_name):
        if self.topic_exists(topic_name=topic_name):
            response = self.__admin.delete_topics([topic_name])
            return response

    def write_data(self, topic_name: str, messages: DataFrame):
        for message in json.loads(messages.to_json(orient='records', date_format="iso")):
            response = self.__producer.send(topic=topic_name, value=message)



import json
import uuid
from datetime import datetime

from sqlalchemy import Row

from pdip.json.encoders.date_time_encoder import DateTimeEncoder
from pdip.json.encoders.mutliple_json_encoders import MultipleJsonEncoders
from pdip.json.encoders.uuid_encoder import UUIDEncoder
from pdip.utils import TypeChecker


class BaseConverter(object):
    def __init__(self, cls=None):
        self.mappings = {}
        self.type_checker = TypeChecker()
        if cls is not None:
            self.register(cls)

    def class_mapper(self, d):
        for keys, cls in self.mappings.items():
            if keys.issuperset(d.keys()):  # are all required arguments present?
                return cls(**d)
        else:
            # Raise exception instead of silently returning None
            raise ValueError(f'Unable to find a matching class for object: {d}')

    def register(self, cls):
        instance = cls()
        mapping_data = frozenset(tuple([attr for attr, val in instance.__dict__.items()]))
        self.mappings[mapping_data] = cls
        annotations = self.get_annotations(instance)
        self.register_subclasses(annotations)
        return cls

    def check_uuid(self, obj):
        if isinstance(obj, Row):
            for i, o in enumerate(obj):
                if isinstance(o, uuid.UUID):
                    obj[i] = str(o)
        return obj

    def ToJSON(self, obj):
        obj_dict = None
        if isinstance(obj, dict):
            obj_dict = obj
        elif isinstance(obj, Row):
            obj_dict = obj._asdict()
        else:
            obj_dict = dict(obj)
        encoder = MultipleJsonEncoders(DateTimeEncoder, UUIDEncoder)
        return json.dumps(obj_dict, cls=encoder, indent=4)

    def FromJSON(self, json_str):
        return json.loads(json_str, object_hook=self.class_mapper)

    def get_annotations(self, obj):
        if hasattr(obj, '__annotations__'):
            annotations = obj.__annotations__
            return annotations

    def register_subclasses(self, annotations):
        if annotations is not None and len(annotations) > 0:
            for key in annotations:
                value = annotations[key]
                if value == int:
                    pass
                elif value == str:
                    pass
                elif value == bool:
                    pass
                elif value == datetime:
                    pass
                elif value == float:
                    pass
                else:
                    if self.type_checker.is_generic(value):

                        if self.type_checker.is_primitive(value.__args__[0]):
                            pass
                        else:
                            self.register(value.__args__[0])
                            instance = value.__args__[0]()
                            nested_annotations = self.get_annotations(instance)
                            if nested_annotations is not None:
                                self.register_subclasses(nested_annotations)
                    elif self.type_checker.is_base_generic(value):
                        # TODO:Base generic class
                        print('value type should be a structure of', value.__args__[0])
                    elif self.type_checker.is_class(value):
                        self.register(value)
                        instance = value()
                        nested_annotations = self.get_annotations(instance)
                        if nested_annotations is not None:
                            self.register_subclasses(nested_annotations)
                    else:
                        print('Type not know', value)



from typing import List

from injector import inject

from pdip.integrator.connection.base import ConnectionSourceAdapter
from pdip.integrator.connection.types.sql.base import SqlProvider
from pdip.integrator.integration.domain.base import IntegrationBase


class SqlSourceAdapter(ConnectionSourceAdapter):
    @inject
    def __init__(self,
                 provider: SqlProvider,
                 ):
        self.provider = provider

    def get_source_data_count(self, integration: IntegrationBase) -> int:
        source_context = self.provider.get_context_by_config(
            config=integration.SourceConnections.Sql.Connection)
        if integration.SourceConnections.Sql.Query is None or integration.SourceConnections.Sql.Query == '':
            schema = integration.SourceConnections.Sql.Schema
            table = integration.SourceConnections.Sql.ObjectName
            source_columns = integration.SourceConnections.Columns
            query = source_context.dialect.prepare_select_query(schema=schema, table=table, columns=source_columns)
            data_count = source_context.get_count_for_query(query=query)
        else:
            data_count = source_context.get_table_count(
                schema=integration.SourceConnections.Sql.Schema,
                table=integration.SourceConnections.Sql.ObjectName
            )
        return data_count

    def get_source_data(self, integration: IntegrationBase) -> List[any]:
        source_context = self.provider.get_context_by_config(
            config=integration.SourceConnections.Sql.Connection)
        query = integration.SourceConnections.Sql.Query
        if integration.SourceConnections.Sql.Query is None or integration.SourceConnections.Sql.Query == '':
            schema = integration.SourceConnections.Sql.Schema
            table = integration.SourceConnections.Sql.ObjectName
            source_columns = integration.SourceConnections.Columns
            query = source_context.dialect.prepare_select_query(schema=schema, table=table, columns=source_columns)
        data = source_context.get_data_for_query(query=query)
        return data

    def get_iterator(self, integration: IntegrationBase, limit):
        source_context = self.provider.get_context_by_config(
            config=integration.SourceConnections.Sql.Connection)
        query = integration.SourceConnections.Sql.Query
        if integration.SourceConnections.Sql.Query is None or integration.SourceConnections.Sql.Query == '':
            schema = integration.SourceConnections.Sql.Schema
            table = integration.SourceConnections.Sql.ObjectName
            source_columns = integration.SourceConnections.Columns
            query = source_context.dialect.prepare_select_query(schema=schema, table=table, columns=source_columns)
        iterator = source_context.get_iterator_for_query(query, limit)
        return iterator

    def get_source_data_with_paging(self, integration: IntegrationBase, start, end) -> List[any]:
        source_context = self.provider.get_context_by_config(
            config=integration.SourceConnections.Sql.Connection)
        query = integration.SourceConnections.Sql.Query
        if integration.SourceConnections.Sql.Query is None or integration.SourceConnections.Sql.Query == '':
            schema = integration.SourceConnections.Sql.Schema
            table = integration.SourceConnections.Sql.ObjectName
            source_columns = integration.SourceConnections.Columns
            query = source_context.dialect.prepare_select_query(schema=schema, table=table, columns=source_columns)

        data = source_context.get_data_with_paging(
            query=query,
            start=start,
            end=end
        )
        return data



from typing import List

from injector import inject

from pdip.integrator.connection.base import ConnectionSourceAdapter
from pdip.integrator.connection.types.bigdata.base import BigDataProvider
from pdip.integrator.integration.domain.base import IntegrationBase


class BigDataSourceAdapter(ConnectionSourceAdapter):
    @inject
    def __init__(self,
                 provider: BigDataProvider,
                 ):
        self.provider = provider

    def get_source_data_count(self, integration: IntegrationBase) -> int:
        source_context = self.provider.get_context_by_config(
            config=integration.SourceConnections.BigData.Connection)
        query = integration.SourceConnections.BigData.Query
        if integration.SourceConnections.BigData.Query is None or integration.SourceConnections.BigData.Query == '':
            schema = integration.SourceConnections.BigData.Schema
            table = integration.SourceConnections.BigData.ObjectName
            if schema is None or schema == '' or table is None or table == '':
                raise Exception(f"Source Schema and Table required. {schema}.{table}")
            query = source_context.dialect.get_table_select_query(selected_rows='*', schema=schema, table=table)
        data_count = source_context.get_table_count(query=query)
        return data_count

    def get_source_data(self, integration: IntegrationBase) -> List[any]:
        source_context = self.provider.get_context_by_config(
            config=integration.SourceConnections.BigData.Connection)
        query = integration.SourceConnections.BigData.Query
        if integration.SourceConnections.BigData.Query is None or integration.SourceConnections.BigData.Query == '':
            schema = integration.SourceConnections.BigData.Schema
            table = integration.SourceConnections.BigData.ObjectName
            if schema is None or schema == '' or table is None or table == '':
                raise Exception(f"Source Schema and Table required. {schema}.{table}")
            query = source_context.dialect.get_table_select_query(selected_rows='*', schema=schema, table=table)
        data = source_context.get_table_data(query=query)
        return data

    def get_iterator(self, integration: IntegrationBase, limit):
        source_context = self.provider.get_context_by_config(
            config=integration.SourceConnections.BigData.Connection)
        query = integration.SourceConnections.BigData.Query
        if integration.SourceConnections.BigData.Query is None or integration.SourceConnections.BigData.Query == '':
            schema = integration.SourceConnections.BigData.Schema
            table = integration.SourceConnections.BigData.ObjectName
            if schema is None or schema == '' or table is None or table == '':
                raise Exception(f"Source Schema and Table required. {schema}.{table}")
            source_columns = integration.SourceConnections.Columns
            if source_columns is not None and len(source_columns) > 0:
                source_column_rows = [column.Name for column in source_columns]
                columns_query = ",".join(source_column_rows)
                query = source_context.dialect.get_table_select_query(selected_rows=columns_query, schema=schema,
                                                                      table=table)
            else:
                query = source_context.dialect.get_table_select_query(selected_rows='*', schema=schema, table=table)
        iterator = source_context.get_iterator(query, limit)
        return iterator

    def get_source_data_with_paging(self, integration: IntegrationBase, start, end) -> List[any]:
        source_context = self.provider.get_context_by_config(
            config=integration.SourceConnections.BigData.Connection)
        query = integration.SourceConnections.BigData.Query
        if integration.SourceConnections.BigData.Query is None or integration.SourceConnections.BigData.Query == '':
            schema = integration.SourceConnections.BigData.Schema
            table = integration.SourceConnections.BigData.ObjectName
            if schema is None or schema == '' or table is None or table == '':
                raise Exception(f"Source Schema and Table required. {schema}.{table}")
            query = source_context.dialect.get_table_select_query(selected_rows='*', schema=schema, table=table)
        data = source_context.get_table_data_with_paging(
            query=query,
            start=start,
            end=end
        )
        return data



import os

from injector import inject

from ..configuration.models.application import ApplicationConfig
from ..dependency import IScoped


class FileManager(IScoped):
    @inject
    def __init__(self,
                 application_config: ApplicationConfig
                 ):
        self.application_config = application_config

    def check_path_exist(self, folder, file):
        path = os.path.join(self.application_config.root_directory, folder, file)
        return os.path.exists(path)

    def create_folder_if_not_exist(self, folder):
        os.path.join(self.application_config.root_directory, folder)
        if not os.path.exists(folder):
            os.makedirs(folder)

    def create_file(self, folder, file_name, content, file_extension='.py'):
        self.create_folder_if_not_exist(folder)
        file_path = '/'.join([folder, f'{file_name}{file_extension}'])

        with open(file_path, "w") as outfile:
            outfile.write(content.strip() + '\n')



import uuid
from datetime import datetime
from typing import Generic, List, TypeVar, Type

from injector import inject
from sqlalchemy.orm import Query

from ..base.database_session_manager import DatabaseSessionManager

T = TypeVar('T')


class Repository(Generic[T]):
    @inject
    def __init__(self, repository_type: Type[T], database_session_manager: DatabaseSessionManager):
        self.database_session_manager: DatabaseSessionManager = database_session_manager
        self.type = repository_type

    @property
    def table(self):
        return self.database_session_manager.session.query(self.type)

    def first(self, **kwargs) -> T:
        query: Query = self.table.filter_by(**kwargs)
        return query.first()

    def filter_by(self, **kwargs) -> List[T]:
        return self.table.filter_by(**kwargs)

    def get(self) -> List[T]:
        return self.table.all()

    def get_by_id(self, id: int) -> T:
        return self.table.filter_by(Id=id).first()

    def insert(self, entity: T):

        if entity.TenantId is None:
            if self.database_session_manager.engine.dialect.name == 'postgresql':
                entity.TenantId = uuid.UUID("00000000-0000-0000-0000-000000000000")
            else:
                entity.TenantId = str(uuid.UUID("00000000-0000-0000-0000-000000000000"))


        entity.CreateUserTime = datetime.now()
        if self.database_session_manager.engine.dialect.name == 'postgresql':
            entity.CreateUserId = uuid.UUID("00000000-0000-0000-0000-000000000000")
        else:
            entity.CreateUserId = str(uuid.UUID("00000000-0000-0000-0000-000000000000"))
        self.database_session_manager.session.add(entity)

    def update(self, entity: T):
        entity.UpdateUserTime = datetime.now()
        if self.database_session_manager.engine.dialect.name == 'postgresql':
            entity.UpdateUserId = uuid.UUID("00000000-0000-0000-0000-000000000000")
        else:
            entity.UpdateUserId = str(uuid.UUID("00000000-0000-0000-0000-000000000000"))

    def delete_by_id(self, id: int):
        entity = self.get_by_id(id)
        entity.GcRecId = uuid.uuid4()
        entity.UpdateUserTime = datetime.now()
        if self.database_session_manager.engine.dialect.name == 'postgresql':
            entity.UpdateUserId = uuid.UUID("00000000-0000-0000-0000-000000000000")
        else:
            entity.UpdateUserId = str(uuid.UUID("00000000-0000-0000-0000-000000000000"))

    def delete(self, entity: T):
        entity.GcRecId = uuid.uuid4()
        entity.UpdateUserTime = datetime.now()
        if self.database_session_manager.engine.dialect.name == 'postgresql':
            entity.UpdateUserId = uuid.UUID("00000000-0000-0000-0000-000000000000")
        else:
            entity.UpdateUserId = str(uuid.UUID("00000000-0000-0000-0000-000000000000"))

    def commit(self):
        self.database_session_manager.commit()



import uuid
from datetime import datetime

from sqlalchemy import DateTime, Column
from sqlalchemy.ext.declarative import declared_attr

from .entity_base import EntityBase
from .types import GUID


class Entity(EntityBase):
    __allow_unmapped__ = True
    
    Id = Column(
        GUID(),
        primary_key=True,
        default=lambda: str(uuid.uuid4())
    )

    @declared_attr
    def CreateUserId(cls):
        return Column(GUID(), index=False, unique=False, nullable=False,
                      default=lambda: uuid.UUID("00000000-0000-0000-0000-000000000000"))

    @declared_attr
    def CreateUserTime(cls):
        return Column(DateTime, index=False, unique=False, nullable=False, default=lambda: datetime.now)

    @declared_attr
    def UpdateUserId(cls):
        return Column(GUID(), index=False, unique=False, nullable=True)

    @declared_attr
    def UpdateUserTime(cls):
        return Column(DateTime, index=False, unique=False, nullable=True)

    @declared_attr
    def TenantId(cls):
        return Column(GUID(), index=False, unique=False, nullable=False)

    @declared_attr
    def GcRecId(cls):
        return Column(GUID(), index=False, unique=False, nullable=True)



from injector import inject

from .web_service_context import WebServiceContext
from .web_service_policy import WebServicePolicy
from ....domain.authentication.basic import ConnectionBasicAuthentication
from ....domain.enums import ConnectorTypes, ConnectionTypes
from ....domain.server.base import ConnectionServer
from ....domain.types.webservice.configuration.base import WebServiceConnectionConfiguration
from ....domain.types.webservice.configuration.soap import SoapConfiguration
from ......dependency import IScoped


class WebServiceProvider(IScoped):
    @inject
    def __init__(self):
        pass

    def __initialize_context(self, config: WebServiceConnectionConfiguration):
        policy = WebServicePolicy(config=config)
        context = WebServiceContext(policy=policy)
        return context

    def get_context_by_config(self, config: WebServiceConnectionConfiguration) -> WebServiceContext:
        return self.__initialize_context(config=config)

    def get_context(
            self,
            connector_type: ConnectorTypes,
            host: str, port: int,
            user: str, password: str,
            wsdl: str,
            ssl: bool = False
    ) -> WebServiceContext:
        """
        Creating Context
        """
        if connector_type == connector_type.Soap:
            config = WebServiceConnectionConfiguration(
                ConnectionType=ConnectionTypes.WebService,
                ConnectorType=ConnectorTypes.Soap,
                Server=ConnectionServer(Host=host, Port=port),
                BasicAuthentication=ConnectionBasicAuthentication(User=user,
                                                                  Password=password),
                Soap=SoapConfiguration(Wsdl=wsdl),
                Ssl=ssl
            )
        else:
            raise Exception(f"{connector_type.name} connector type not supported")

        return self.__initialize_context(config=config)



from injector import inject
from sqlalchemy import desc, text

from ...api.request_parameter import OrderByParameter
from ...dependency import IScoped
from ...utils import ModuleFinder


class OrderBySpecification(IScoped):
    @inject
    def __init__(self, module_finder: ModuleFinder):
        self.default_order = 'asc'
        self.module_finder = module_finder

    def specify(self, order_by_parameter: OrderByParameter):

        if not hasattr(order_by_parameter, 'OrderBy') and hasattr(order_by_parameter, 'Order'):
            return None
        if order_by_parameter.OrderBy is not None and order_by_parameter.OrderBy != '':
            if order_by_parameter.Order is None or order_by_parameter.Order == '':
                order = self.default_order
            else:
                order = order_by_parameter.Order
            split = order_by_parameter.OrderBy.split(".")
            if len(split) == 1:
                return text(f'"{order_by_parameter.OrderBy}" {order_by_parameter.Order}')
            if len(split) == 2:
                class_name, attr_name = split
                module = self.module_finder.get_module(class_name)
                class_type = getattr(module, class_name)
                order_by = getattr(class_type, attr_name)

                if order == 'desc':
                    return desc(order_by)
                else:
                    return order_by
            else:
                return None



from flask import Flask
from flask_cors import CORS
from flask_restx import Api
from injector import inject
from werkzeug.exceptions import HTTPException

from ..handlers import ErrorHandlers
from ..handlers import RequestHandler
from ...configuration.models.api import ApiConfig
from ...configuration.models.application import ApplicationConfig
from ...dependency import ISingleton
from ...exceptions import OperationalException


class FlaskAppWrapper(ISingleton):
    @inject
    def __init__(self,
                 request_handler: RequestHandler,
                 app: Flask,
                 api: Api,
                 api_config: ApiConfig,
                 error_handlers: ErrorHandlers,
                 application_config: ApplicationConfig,
                 ):
        self.application_config = application_config
        self.api = api
        self.request_handler = request_handler
        self.app = app
        self.api_config = api_config
        self.error_handlers = error_handlers
        # Application create operations
        self.create_application()
        # self.find_resources()

    # Application flask configurations and api endpoint registration
    def create_application(self):
        CORS(self.app, resources={r"/api/*": {"origins": "*"}})
        self.register_error_handlers()

    def register_error_handlers(self):
        self.api.errorhandler(OperationalException)(self.error_handlers.handle_operational_exception)
        self.app.register_error_handler(Exception, self.error_handlers.handle_exception)
        self.app.register_error_handler(HTTPException, self.error_handlers.handle_http_exception)
        self.app.after_request(self.request_handler.after_request)

    def run(self):
        self.app.run(debug=self.api_config.is_debug, host='0.0.0.0', port=self.api_config.port)

    def test_client(self):
        return self.app.test_client()



from injector import inject

from ..base import ParallelSessionBroker
from .........dependency import IScoped
from .........logging.loggers.console import ConsoleLogger


class ParallelSessionBrokerFactory(IScoped):
    @inject
    def __init__(self, logger: ConsoleLogger):
        self.logger = logger

    def get(self):
        process_manager = ParallelSessionBroker(logger=self.logger)
        return process_manager



import importlib

from injector import inject

from .big_data_connector import BigDataConnector
from .big_data_dialect import BigDataDialect
from ....domain.enums import ConnectorTypes
from ....domain.types.bigdata.configuration.base import BigDataConnectionConfiguration


class BigDataPolicy:
    @inject
    def __init__(self, config: BigDataConnectionConfiguration):
        self.config = config
        self.connector: BigDataConnector = None
        self.connector_name = None
        connector_base_module = "pdip.integrator.connection.types.bigdata.connectors"
        dialect_base_module = "pdip.integrator.connection.types.bigdata.dialects"
        if self.config.ConnectorType == ConnectorTypes.Impala:
            namespace = "impala"
            connector_name = "ImpalaConnector"
            dialect_name = "ImpalaDialect"
        else:
            raise Exception("Connector type not found")
        module = importlib.import_module(".".join([connector_base_module, namespace]))
        connector_class = getattr(module, connector_name)
        if connector_class is not None:
            self.connector: BigDataConnector = connector_class(self.config)
        module = importlib.import_module(".".join([dialect_base_module, namespace]))
        dialect_class = getattr(module, dialect_name)
        if dialect_class is not None:
            self.dialect: BigDataDialect = dialect_class(self.connector)



from dataclasses import dataclass

from ..base.base_config import BaseConfig


@dataclass
class DatabaseConfig(BaseConfig):
    type: str = None
    connection_string: str = None
    driver: str = None
    host: str = None
    port: int = None
    sid: str = None
    service_name: str = None
    database: str = None
    user: str = None
    password: str = None
    application_name: str = None
    execution_options: str = None



import multiprocessing
from multiprocessing.context import Process
from multiprocessing.managers import SyncManager
from queue import Queue
from typing import List

from .subprocess import Subprocess
from ..models import ProcessInfo
from ..models import ProcessTask
from ...dependency.container import DependencyContainer
from ...logging.loggers.console import ConsoleLogger


class ProcessManager:
    def __init__(
            self,
            logger: ConsoleLogger
    ):
        self._manager: SyncManager = None
        self._process_queue: Queue = None
        self._process_result_queue: Queue = None
        self._processes: List[ProcessInfo] = None
        self._process_tasks: List[ProcessTask] = None
        self.logger = logger
        self.number_of_cores = multiprocessing.cpu_count()

    def __del__(self):
        self.__finish_all_processes()
        if self._manager is not None:
            self._manager.shutdown()

    def get_manager(self):
        if self._manager is None:
            self._manager = multiprocessing.Manager()
        return self._manager

    def create_queue(self) -> Queue:
        return self.get_manager().Queue()

    def start_processes(self, target_method, kwargs, process_count=1):
        self.__configure_process()
        # Initiate the worker processes
        for i in range(process_count):
            # Set process name
            sub_process_id = i + 1
            # Create the process, and connect it to the worker function
            process = self.__start_process(
                sub_process_id=sub_process_id,
                process_queue=self._process_queue,
                process_result_queue=self._process_result_queue,
                target_method=target_method,
                kwargs=kwargs)
            # Add new process to the list of processes
            process_data = ProcessInfo(
                Process=process, SubProcessId=sub_process_id)
            self._processes.append(process_data)
            process_task = ProcessTask(
                SubProcessId=sub_process_id, IsFinished=False)
            self._process_queue.put(process_task)
            self._process_tasks.append(process_task)
            # Start the process

    def get_results(self) -> List[ProcessTask]:
        self.__check_processes()
        return self._process_tasks

    def __configure_process(self):
        self._process_queue = self.create_queue()
        self._process_result_queue = self.create_queue()
        self._processes: List[ProcessInfo] = []
        self._process_tasks: List[ProcessTask] = []

    def __start_process(self,
                        sub_process_id: int,
                        process_queue: Queue,
                        process_result_queue: Queue,
                        target_method: any,
                        kwargs: any) -> Process:
        # Create the process, and connect it to the worker function
        if hasattr(DependencyContainer, 'Instance') and DependencyContainer.Instance is not None:
            root_directory = DependencyContainer.Instance.root_directory
            initialize_container = True
        else:
            root_directory = None
            initialize_container = False
        new_process = Process(
            target=self.start_subprocess,
            args=(
                sub_process_id, root_directory, initialize_container, process_queue, process_result_queue,
                target_method, kwargs))

        new_process.start()
        return new_process

    @staticmethod
    def start_subprocess(sub_process_id: int, root_directory: str, initialize_container: bool,
                         process_queue: Queue,
                         process_result_queue: Queue, target_method,
                         kwargs):
        Subprocess(
            root_directory=root_directory,
            initialize_container=initialize_container
        ) \
            .start(
            sub_process_id=sub_process_id,
            process_queue=process_queue,
            process_result_queue=process_result_queue,
            target_method=target_method,
            kwargs=kwargs
        )

    def __check_processes(self):
        # Read calculation results
        while True:
            # Read result
            process_task: ProcessTask = self._process_result_queue.get()
            # Have a look at the results
            if process_task.IsFinished:
                # Process has finished
                if process_task.State == 4:
                    self.__finish_all_processes()
                    for task in self._process_tasks:
                        if task.SubProcessId == process_task.SubProcessId:
                            task.Result = process_task.Result
                            task.State = process_task.State
                            task.Message = process_task.Message
                            task.IsFinished = process_task.IsFinished
                            task.Exception = process_task.Exception
                            task.Traceback = process_task.Traceback
                            task.Data = process_task.Data
                    break
                for process_data in self._processes:
                    if process_data.SubProcessId == process_task.SubProcessId:
                        process_data.IsFinished = True
                for task in self._process_tasks:
                    if task.SubProcessId == process_task.SubProcessId:
                        task.Result = process_task.Result
                        task.State = process_task.State
                        task.Message = process_task.Message
                        task.IsFinished = process_task.IsFinished
                        task.Exception = process_task.Exception
                        task.Traceback = process_task.Traceback
                        task.Data = process_task.Data
                self.__check_unfinished_processes()
                if self.__check_all_processes_finished():
                    break

    def __finish_all_processes(self):
        if self._processes is not None:
            for process in self._processes:
                if process.Process.is_alive():
                    process_task = ProcessTask(
                        SubProcessId=process.SubProcessId, IsFinished=False)
                    self._process_queue.put(process_task)
                    self.logger.info(f"Process will terminate. SubProcessId:{process.SubProcessId}")
                    process.IsFinished = True
                    process.Process.terminate()

    def __check_unfinished_processes(self):

        for process_data in self._processes:
            if not process_data.IsFinished and not process_data.Process.is_alive():
                self.logger.info(f"Unfinished process found. SubProcessId:{process_data.SubProcessId}")
                process_data.IsFinished = True

    def __check_all_processes_finished(self):
        check_finish = True
        unfinished_process_list = []
        for process_data in self._processes:
            if not process_data.IsFinished:
                check_finish = False
                unfinished_process_list.append(str(process_data.SubProcessId))
        process_join = ",".join(unfinished_process_list)
        self.logger.info(f"All processes are expected to end . SubProcessId:{process_join}")
        return check_finish



from typing import List

from injector import inject
from pandas import DataFrame

from pdip.integrator.connection.base import ConnectionTargetAdapter
from pdip.integrator.connection.types.sql.base import SqlProvider
from pdip.integrator.integration.domain.base import IntegrationBase


class SqlTargetAdapter(ConnectionTargetAdapter):
    @inject
    def __init__(self,
                 provider: SqlProvider,
                 ):
        self.provider = provider

    def clear_data(self, integration: IntegrationBase) -> int:
        target_context = self.provider.get_context_by_config(
            config=integration.TargetConnections.Sql.Connection)
        truncate_affected_rowcount = target_context.truncate_table(schema=integration.TargetConnections.Sql.Schema,
                                                                   table=integration.TargetConnections.Sql.ObjectName)
        return truncate_affected_rowcount

    def prepare_data(self, target_context, integration: IntegrationBase, source_data: any) -> List[any]:
        columns = integration.SourceConnections.Columns
        if columns is not None:
            source_columns = [(column.Name) for column in columns]
        elif columns is None:
            source_columns = list(source_data[0].keys())
        if isinstance(source_data, DataFrame):
            data = source_data[source_columns]
            prepared_data = data.values.tolist()
        else:
            prepared_data = target_context.prepare_insert_row(
                data=source_data,
                columns=source_columns
            )
        return prepared_data

    def write_data(self, integration: IntegrationBase, source_data: List[any]) -> int:
        if source_data is not None and len(source_data) > 0:
            target_context = self.provider.get_context_by_config(
                config=integration.TargetConnections.Sql.Connection
            )
            prepared_data = self.prepare_data(
                target_context=target_context,
                integration=integration,
                source_data=source_data
            )
            source_columns = None
            if integration.SourceConnections.Columns is not None and len(integration.SourceConnections.Columns) > 0:
                source_columns = [column.Name for column in integration.SourceConnections.Columns]
            target_columns = None
            if integration.TargetConnections.Columns is not None and len(integration.TargetConnections.Columns) > 0:
                target_columns = [column.Name for column in integration.TargetConnections.Columns]
            prepared_target_query = target_context.get_target_query(
                query=integration.TargetConnections.Sql.Query,
                source_columns=source_columns,
                target_columns=target_columns,
                schema=integration.TargetConnections.Sql.Schema,
                table=integration.TargetConnections.Sql.ObjectName,
                source_column_count=len(prepared_data[0])
            )
            affected_row_count = target_context.execute_many(
                query=prepared_target_query,
                data=prepared_data
            )
            return affected_row_count
        else:
            return 0

    def do_target_operation(self, integration: IntegrationBase) -> int:
        target_context = self.provider.get_context_by_config(
            config=integration.TargetConnections.Sql.Connection)

        affected_rowcount = target_context.execute(query=integration.TargetConnections.Sql.Query)
        return affected_rowcount



import multiprocessing
from queue import Queue

from injector import inject

from ..operation.source import SourceReadOperation
from ..operation.target import TargetWriteOperation
from ..threading.base import ParallelSessionBroker
from ...base import IntegrationSourceToTargetExecuteStrategy
from .......connection.factories import ConnectionSourceAdapterFactory, ConnectionTargetAdapterFactory
from .......domain.enums.events import EVENT_LOG
from .......operation.domain import OperationIntegrationBase
from .......pubsub.base import ChannelQueue
from .......pubsub.domain import TaskMessage
from .......pubsub.publisher import Publisher
from ........dependency import IScoped
from ........dependency.container import DependencyContainer


class ParallelThreadIntegrationExecute(IntegrationSourceToTargetExecuteStrategy, IScoped):
    @inject
    def __init__(
            self,
            connection_source_adapter_factory: ConnectionSourceAdapterFactory,
            connection_target_adapter_factory: ConnectionTargetAdapterFactory
    ):
        self.connection_source_adapter_factory = connection_source_adapter_factory
        self.connection_target_adapter_factory = connection_target_adapter_factory

    def execute(
            self,
            operation_integration: OperationIntegrationBase,
            channel: ChannelQueue
    ) -> int:
        publisher = Publisher(channel=channel)
        try:
            try:
                manager = multiprocessing.Manager()
                source_data_process_manager = ParallelSessionBroker(channel=channel,
                                                                    operation_integration=operation_integration)
                execute_data_process_manager = ParallelSessionBroker(channel=channel,
                                                                     operation_integration=operation_integration)
                data_queue = manager.Queue()
                data_result_queue = manager.Queue()

                self.start_source_data_subprocess(
                    source_data_process_manager=source_data_process_manager,
                    channel=channel,
                    operation_integration=operation_integration,
                    data_queue=data_queue,
                    data_result_queue=data_result_queue
                )

                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"Source data process started"
                        }
                    )
                )

                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"Execute data process will start"
                        }
                    )
                )
                total_row_count = self.start_execute_data_subprocess(
                    execute_data_process_manager=execute_data_process_manager,
                    channel=channel,
                    operation_integration=operation_integration,
                    data_queue=data_queue,
                    data_result_queue=data_result_queue
                )

                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"Execute data process finished"
                        }
                    )
                )
            finally:
                manager.shutdown()
                del source_data_process_manager
                del execute_data_process_manager

            return total_row_count
        except Exception as ex:
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"Integration getting error.",
                        'exception': ex
                    }
                )
            )
            raise

    def start_source_data_subprocess(
            self,
            source_data_process_manager: ParallelSessionBroker,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ):
        source_data_kwargs = {
            "operation_integration": operation_integration,
            "channel": channel,
            "data_queue": data_queue,
            "data_result_queue": data_result_queue,
        }

        source_data_process_manager.start(
            process_count=1,
            target_method=self.start_source_data_process,
            target_kwargs=source_data_kwargs
        )

    def start_execute_data_subprocess(
            self,
            execute_data_process_manager: ParallelSessionBroker,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ) -> int:
        total_row_count = 0
        execute_data_kwargs = {
            "channel": channel,
            "operation_integration": operation_integration,
            "data_queue": data_queue,
            "data_result_queue": data_result_queue,
        }
        execute_data_process_manager.start(
            process_count=operation_integration.ProcessCount,
            target_method=self.start_execute_data_process,
            target_kwargs=execute_data_kwargs
        )
        execute_data_process_results = execute_data_process_manager.get_results()
        for result in execute_data_process_results:
            if result.Exception is not None:
                raise result.Exception
            if result.Result is not None:
                total_row_count = total_row_count + result.Result
        return total_row_count

    @staticmethod
    def start_source_data_process(
            thread_id,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ):
        return DependencyContainer.Instance.get(SourceReadOperation).start(
            thread_id=thread_id,
            channel=channel,
            operation_integration=operation_integration,
            data_queue=data_queue,
            data_result_queue=data_result_queue,
        )

    @staticmethod
    def start_execute_data_process(
            thread_id,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ) -> int:
        return DependencyContainer.Instance.get(TargetWriteOperation).start(
            thread_id=thread_id,
            channel=channel,
            operation_integration=operation_integration,
            data_queue=data_queue,
            data_result_queue=data_result_queue,
        )



import json


class MultipleJsonEncoders:
    """
    Combine multiple JSON encoders
    """

    def __init__(self, *encoders):
        self.encoders = encoders
        self.args = ()
        self.kwargs = {}

    def default(self, obj):
        for encoder in self.encoders:
            try:
                return encoder(*self.args, **self.kwargs).default(obj)
            except TypeError:
                pass
        raise TypeError(f'Object of type {obj.__class__.__name__} is not JSON serializable')

    def __call__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        enc = json.JSONEncoder(*args, **kwargs)
        enc.default = self.default
        return enc



from abc import ABC, abstractmethod

from injector import inject

from pdip.dependency import IScoped
from pdip.integrator.operation.domain import OperationIntegrationBase
from pdip.integrator.pubsub.base import ChannelQueue


class IntegrationSourceToTargetExecuteStrategy(ABC, IScoped):
    @inject
    def __init__(self):
        pass

    @abstractmethod
    def execute(
            self,
            operation_integration: OperationIntegrationBase,
            channel: ChannelQueue
    ) -> int:
        pass



from injector import inject

from .execution_initializer import ExecutionInitializer
from ....execution.domain import ExecutionOperationBase, ExecutionOperationIntegrationBase
from ....operation.domain import OperationBase
from .....dependency import IScoped


class DefaultExecutionInitializer(ExecutionInitializer, IScoped):
    @inject
    def __init__(self):
        pass

    def initialize(
            self,
            operation: OperationBase,
            execution_id: int = None,
            ap_scheduler_job_id: int = None
    ) -> OperationBase:
        execution_operation = ExecutionOperationBase(
            Id=execution_id,
            ApSchedulerJobId=ap_scheduler_job_id,
            OperationId=operation.Id,
            Name=operation.Name,
            Events=[]
        )
        operation.Execution = execution_operation

        for operation_integration in operation.Integrations:
            execution_operation_integration = ExecutionOperationIntegrationBase(
                Id=None,
                OperationId=operation.Id,
                OperationExecutionId=execution_id,
                ApSchedulerJobId=ap_scheduler_job_id,
                OperationIntegrationId=operation_integration.Id,
                Name=operation_integration.Name,
                Events=[]
            )
            operation_integration.Execution = execution_operation_integration
        return operation



from .command_query_base import CommandQueryBase
from .command_query_handler_base import CommandQueryHandlerBase
from .dispatcher import Dispatcher
from .icommand import ICommand
from .icommand_handler import ICommandHandler
from .iquery import IQuery
from .iquery_handler import IQueryHandler



from injector import inject

from ..initializer.integrator import IntegratorInitializerFactory
from ..operation.base import OperationExecution
from ..operation.domain import OperationBase
from ..pubsub.base import MessageBroker
from ...logging.loggers.console import ConsoleLogger


class Integrator:
    @inject
    def __init__(
            self,
            logger: ConsoleLogger,
            integrator_initializer_factory: IntegratorInitializerFactory,
            operation_execution: OperationExecution
    ):
        self.operation_execution = operation_execution
        self.logger = logger
        self.integrator_initializer_factory = integrator_initializer_factory
        self.message_broker = MessageBroker(self.logger)

    def __del__(self):
        self.close()

    def close(self):
        if hasattr(self, 'message_broker'):
            del self.message_broker

    def integrate(
            self,
            operation: any,
            execution_id: int = None,
            ap_scheduler_job_id: int = None
    ):
        if operation is None:
            raise Exception('Operation required')

        if isinstance(operation, OperationBase):
            try:
                operation = self.integrator_initializer_factory \
                    .get() \
                    .initialize(operation=operation,
                                message_broker=self.message_broker,
                                execution_id=execution_id,
                                ap_scheduler_job_id=ap_scheduler_job_id
                                )
                self.operation_execution.start(
                    operation=operation,
                    channel=self.message_broker.get_publish_channel()
                )
            except Exception as ex:
                raise
            finally:
                self.message_broker.join()
                self.close()
        else:
            raise Exception('Operation type is not suitable')

    def subscribe(self, event, callback):
        self.message_broker.subscribe(event, callback)

    def unsubscribe(self, event, callback):
        self.message_broker.unsubscribe(event, callback)



import traceback
from queue import Queue

from injector import inject
from pdip.dependency import IScoped

from pdip.integrator.connection.domain.task import DataQueueTask
from pdip.integrator.connection.factories import ConnectionSourceAdapterFactory
from pdip.integrator.domain.enums.events import EVENT_LOG
from pdip.integrator.operation.domain import OperationIntegrationBase
from pdip.integrator.pubsub.base import ChannelQueue
from pdip.integrator.pubsub.domain import TaskMessage
from pdip.integrator.pubsub.publisher import Publisher


class SourceReadOperation(IScoped):
    @inject
    def __init__(
            self,
            connection_source_adapter_factory: ConnectionSourceAdapterFactory,
    ):
        self.connection_source_adapter_factory = connection_source_adapter_factory

    def start(
            self,
            thread_id,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ):
        publisher = Publisher(channel=channel)
        publisher.publish(
            message=TaskMessage(
                event=EVENT_LOG,
                kwargs={
                    'data': operation_integration,
                    'message': f"Source data operation started on process. thread_id: {thread_id}"
                }
            )
        )
        try:

            source_adapter = self.connection_source_adapter_factory.get_adapter(
                connection_type=operation_integration.Integration.SourceConnections.ConnectionType
            )

            data_count = source_adapter.get_source_data_count(integration=operation_integration.Integration)
            if data_count > 0:
                transmitted_data_count = 0
                limit = operation_integration.Limit
                end = limit
                start = 0
                id = 0
                while True:
                    if end != limit and end - data_count >= limit:
                        break
                    id = id + 1
                    data_queue_task = DataQueueTask(
                        Id=id,
                        Data=None,
                        IsDataFrame=False,
                        Start=start,
                        End=end,
                        Limit=limit,
                        IsFinished=False
                    )
                    data_queue.put(data_queue_task)
                    transmitted_data_count = transmitted_data_count + 1
                    if transmitted_data_count >= operation_integration.ProcessCount:
                        result = data_result_queue.get()
                        if result:
                            transmitted_data_count = transmitted_data_count - 1
                        else:
                            break
                    end += limit
                    start += limit
            for i in range(operation_integration.ProcessCount):
                data_queue_finish_task = DataQueueTask(IsFinished=True)
                data_queue.put(data_queue_finish_task)

            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"Source data operation finished successfully. thread_id: {thread_id}"
                    }
                )
            )
        except Exception as ex:
            exception_traceback = traceback.format_exc()
            for i in range(operation_integration.ProcessCount):
                data_queue_error_task = DataQueueTask(
                    IsFinished=True,
                    Traceback=exception_traceback,
                    Exception=ex
                )
                data_queue.put(data_queue_error_task)

            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"Source data operation finished with error. thread_id: {thread_id}.",
                        'exception': ex

                    }
                )
            )
            raise



from abc import abstractmethod

from ..base import Initializer
from ...operation.domain import OperationBase
from ...pubsub.base import MessageBroker


class IntegratorInitializer(Initializer):
    @abstractmethod
    def initialize(
            self,
            operation: OperationBase,
            message_broker: MessageBroker,
            execution_id: int = None,
            ap_scheduler_job_id: int = None
    ) -> OperationBase:
        pass



from typing import Generic, TypeVar

from .command_query_base import CommandQueryBase

C = TypeVar('C', covariant=True)


class ICommand(Generic[C], CommandQueryBase[C]):
    def __init__(self) -> C:
        pass



from datetime import datetime
from uuid import UUID


class EntityBase:

    def __init__(self,
                 Id: UUID = None,
                 CreateUserId: UUID = None,
                 CreateUserTime: datetime = None,
                 UpdateUserId: UUID = None,
                 UpdateUserTime: datetime = None,
                 TenantId: UUID = None,
                 GcRecId: UUID = None,
                 # RowVersion: bytes = None,
                 *args, **kwargs
                 ):
        super().__init__(*args, **kwargs)
        self.Id = Id
        self.CreateUserId: int = CreateUserId
        self.CreateUserTime: datetime = CreateUserTime
        self.UpdateUserId: int = UpdateUserId
        self.UpdateUserTime: datetime = UpdateUserTime
        self.TenantId: bool = TenantId
        self.GcRecId: bool = GcRecId
        # self.RowVersion: bytes = RowVersion



import glob
import importlib
import os
import sys

from .utils import Utils

class ModuleFinder:
    def __init__(self, root_directory: str, initialize: bool = True):
        self.root_directory = root_directory
        self.running_directory = os.getcwd()
        self.modules = []
        self.get_indexes = lambda module_name, modules: [i for (module, i) in zip(modules, range(len(modules))) if
                                                         module["module_name"] == module_name]
        if initialize:
            self.find_all_modules(self.root_directory)

    def get_file_name(self, file):
        file_splits = Utils.path_split(file)
        file_name = file_splits[len(file_splits) - 1]
        file_name_without_extension = file_name[0:(len(file_name) - 3)]
        return file_name_without_extension

    def find_sub_folders(self, directory):
        for name in os.listdir(directory):
            sub_folder = os.path.join(directory, name)
            if os.path.isdir(sub_folder) and not name.startswith('.') and not name.startswith('_') and name != 'files':
                yield sub_folder
                for folder in self.find_sub_folders(sub_folder):
                    yield folder

    def find_last_parent(self):
        return min([p for p in sys.path if p in self.root_directory], key=len)

    def find_last_parent_module(self):
        last_parent = self.find_last_parent()
        if (last_parent != self.root_directory):
            module_name = self.root_directory.replace(last_parent, '')[
                          1:].replace('\\', '.').replace('/', '.')
            return module_name

    def find_all_modules(self, folder):
        folder_path = os.path.join(folder)
        folders = self.find_sub_folders(folder_path)
        module_base_address = ''
        module_last_parent_address = ''
        if (self.running_directory == self.root_directory):
            last_parent_module_address = self.find_last_parent_module()
            if last_parent_module_address is not None:
                module_last_parent_address = last_parent_module_address
        elif (self.running_directory != self.root_directory):
            module_base_address = self.root_directory.lower().replace(self.running_directory.lower(), '')[
                                  1:].replace('\\', '.').replace('/', '.')

        for folder in folders:
            files = glob.glob(folder + '/*.py')
            for file in files:
                module = {}
                file_name = self.get_file_name(file=file)
                module_path = os.path.join(folder, file_name)
                module_address = module_path.replace(self.root_directory, '')[
                                 1:].replace('\\', '.').replace('/', '.')
                module['module_name'] = file_name
                module['file_path'] = file
                module['module_path'] = module_path
                module['module_address'] = module_address
                module['module_base_address'] = module_base_address
                module['module_parent_address'] = '.'.join(module_address.split('.')[:-1])
                module['module_last_parent_address'] = module_last_parent_address
                self.modules.append(module)

    def import_modules_by_name_ends_with(self, name):
        for module in self.modules:
            if (module['module_name'].endswith(name)):
                module_address = module["module_address"]
                if module_address not in sys.modules:
                    importlib.import_module(module_address)

    # checking existing module for duplicate importing
    def check_module_existing(self, module_address: str):
        for k in sys.modules.keys():
            if k.endswith(module_address):
                return True
        else:
            return False

    def import_modules(self, included_modules=None, excluded_modules=None):
        try:
            for module in self.modules:
                root_path_dir = os.path.join(
                    self.root_directory, '')
                base_module_folder = \
                    module['module_path'].replace(root_path_dir, '')
                if ((excluded_modules is None) or (
                        not any(base_module_folder.startswith(item) for item in excluded_modules))) and (
                        (included_modules is None) or (included_modules is not None and any(
                    base_module_folder.startswith(item) for item in included_modules))):
                    module_address = module["module_address"]
                    module_last_parent_address = module["module_last_parent_address"]

                    if not self.check_module_existing(module_address=module_address):
                        try:
                            module_to_be_added = module_address
                            if module_last_parent_address is not None and module_last_parent_address != '':
                                module_to_be_added = '.'.join([module_last_parent_address, module_address])
                            importlib.import_module(module_to_be_added)
                        except ModuleNotFoundError as ex:
                            # print("ModuleNotFoundError:" + str(ex))
                            module_base_address = module["module_base_address"]
                            module_to_be_added = '.'.join([module_base_address, module_address])
                            try:
                                importlib.import_module(module_to_be_added)
                            except KeyError:
                                pass
                        except Exception as ex:
                            raise

        except:
            self.cleanup()
            raise

    def get_module(self, name_of_module):
        indexes = self.get_indexes(name_of_module, self.modules)
        if indexes is not None and len(indexes) > 0:
            if len(indexes) == 1:
                module_address = self.modules[indexes[0]]["module_address"]
                module = importlib.import_module(module_address)
                return module
        else:
            raise Exception("Modules not found")

    def cleanup(self):
        for module in self.modules:
            module_names = [k for k in sys.modules.keys() if
                            module["module_address"] in k or module["module_parent_address"] in k]
            for module_name in module_names:
                if module_name in sys.modules:
                    del sys.modules[module_name]



import json
from functools import wraps

from .endpoint_wrapper import EndpointWrapper
from ...utils import TypeChecker


class Endpoint:
    def __init__(self, function, namespace, endpoint_wrapper: EndpointWrapper):
        self.endpoint_wrapper = endpoint_wrapper
        self.namespace = namespace
        self.function = function

    def __call__(self, *args, **kwargs):
        input_type, input_name = self.find_input_type()
        req = None
        if input_type is not None and input_name is not None:

            if TypeChecker().is_class(input_type):
                if self.function.__name__ == 'get':
                    req = self.endpoint_wrapper.get_request_from_parser(input_type)
                else:
                    req = self.endpoint_wrapper.get_request_from_body(input_type)
            else:
                req = self.endpoint_wrapper.get_request_from_parser_for_primitive(input_name, input_type)
        if len(self.input_types()[0]) == 1:
            res = self.function(args[0], req, **kwargs)
        else:
            res = self.function(args[0], **kwargs)

        if self.return_type() is not None:
            if hasattr(res, 'to_dict'):
                result = json.loads(json.dumps(res.to_dict(), default=self.endpoint_wrapper.date_converter))
            else:
                result = res
            endpoint_response = self.endpoint_wrapper.get_response(result=result)
            return endpoint_response
        else:
            endpoint_response = self.endpoint_wrapper.get_response(result=res)
            return endpoint_response

    def input_type_names(self):
        self.function.__annotations__.keys()
        return self.function.__annotations__.keys().remove("return")

    def input_types(self):
        input_annotations = self.function.__annotations__
        input_argument_types = []
        input_argument_names = []
        for annotation_key in input_annotations.keys():
            if annotation_key != "return":
                argument_type = input_annotations[annotation_key]
                input_argument_types.append(argument_type)
                input_argument_names.append(annotation_key)
        return input_argument_types, input_argument_names

    def return_type(self):
        if "return" in self.function.__annotations__:
            return self.function.__annotations__["return"]
        else:
            return None

    def find_input_type(self):
        input_types, input_names = self.input_types()
        if input_types == None or len(input_types) == 0:
            return None, None
        elif len(input_types) == 1:
            return input_types[0], input_names[0]
        else:
            return None, None

    def expect_inputs(self):
        input_type, input_name = self.find_input_type()
        expect_model = None
        if input_type is not None and input_name is not None:
            if TypeChecker().is_class(input_type):
                if self.function.__name__ == 'get':
                    expect_model = self.endpoint_wrapper.request_parser(input_type)
                else:
                    expect_model = self.endpoint_wrapper.request_model(input_type)
            else:

                expect_model = self.endpoint_wrapper.create_parser(input_name, input_type)

        return expect_model

    def marshal_with_fields(self):
        if self.return_type() is None:
            fields = self.endpoint_wrapper.BaseModel
        else:
            fields = self.endpoint_wrapper.response_model(self.return_type())
        return fields


def endpoint(api, namespace):
    def decorator(function):
        function.decorated = True

        def instance() -> Endpoint:
            endpoint_wrapper = EndpointWrapper(api=api)
            _instance = Endpoint(function=function, namespace=namespace, endpoint_wrapper=endpoint_wrapper)
            return _instance

        def expected_inputs():
            result = instance().expect_inputs()
            return result

        def marshal_with_fields():
            result = instance().marshal_with_fields()
            return result

        @wraps(function)
        @namespace.expect(expected_inputs(), validate=True)
        @namespace.marshal_with(marshal_with_fields())
        def wrapper(*args, **kwargs):
            return instance().__call__(*args, **kwargs)

        return wrapper

    return decorator



from dataclasses import dataclass

from ..base.base_config import BaseConfig


@dataclass
class ApplicationConfig(BaseConfig):
    root_directory: str = None
    name: str = None
    environment: str = None
    hostname: str = None
    secret_key: str = None



from sqlalchemy import inspect

from ...base import SqlDialect, SqlConnector


class OracleDialect(SqlDialect):
    def __init__(self, connector: SqlConnector):
        self.connector = connector

    @property
    def inspector(self):
        engine = self.connector.get_engine()
        return inspect(engine)

    @property
    def indexer(self) -> str:
        return ':{index}'

    @property
    def quotation_mark(self) -> str:
        return '"'

    def mark_to_object(self, o) -> str:
        return f'{self.quotation_mark}{o}{self.quotation_mark}'

    def get_table_count_query(self, schema, table):
        return f'SELECT COUNT(*) {self.mark_to_object("COUNT")} FROM {self.mark_to_object(schema)}.{self.mark_to_object(table)}'

    def get_count_query(self, query):
        return f'SELECT COUNT(*) {self.mark_to_object("COUNT")} FROM ({query})'

    def get_table_select_query(self, schema, table, selected_rows):
        return f'SELECT {selected_rows} FROM {self.mark_to_object(schema)}.{self.mark_to_object(table)}'

    def get_select_query(self, query):
        return f"SELECT * FROM ({query}) base_query"

    def get_paging_query(self, query, start, end):
        return f'''
SELECT * FROM
(
    SELECT base_query.*, rownum "row_number"
    FROM
    (
        {query}
    ) base_query
    WHERE rownum <= {end}
)
WHERE "row_number" > {start}
'''

    def prepare_select_query(self, schema, table, columns=None):
        if schema is None or schema == '' or table is None or table == '':
            raise Exception(f"Source Schema and Table required. {schema}.{table}")
        if columns is not None and len(columns) > 0:
            source_column_rows = [self.mark_to_object(column.Name) for column in columns]
            columns_query = ",".join(source_column_rows)
            query = self.get_table_select_query(selected_rows=columns_query, schema=schema,
                                                table=table)
        else:
            query = self.get_table_select_query(selected_rows='*', schema=schema, table=table)
        return query

    def get_insert_values_query(self, schema, table, values_query):
        return f'insert into {self.mark_to_object(schema)}.{self.mark_to_object(table)} values({values_query})'

    def get_insert_query(self, schema, table, columns_query, values_query):
        return f'insert into {self.mark_to_object(schema)}.{self.mark_to_object(table)}({columns_query}) values({values_query})'

    def prepare_insert_query(self, schema, table, columns=None, source_column_count: int = None):
        if schema is None or schema == '' or table is None or table == '':
            raise Exception(f"Schema and table required. {schema}.{table}")
        indexer_array = []
        for index in range(source_column_count):
            column_indexer = self.indexer.format(index=index)
            indexer_array.append(column_indexer)
        values_query = ','.join(indexer_array)
        query = self.get_insert_values_query(
            values_query=values_query,
            schema=schema,
            table=table
        )
        return query

    def get_create_table_query(self, schema, table, columns, if_exists=None):
        query = f'''CREATE TABLE {self.mark_to_object(schema)}.{self.mark_to_object(table)} ('''
        column_queries = []
        for column in columns:
            name = f'{self.mark_to_object(column.Name)}'
            column_type = column.Type
            nullable = '' if not column.Nullable else "NULL"
            column_query = " ".join([name, column_type, nullable])
            column_queries.append(column_query)
        query += ",\n".join(column_queries)
        query += f''')'''

        if if_exists == 'DoNothing':
            query = f'''declare
v_sql LONG;
begin

v_sql:='{query}';
execute immediate v_sql;

EXCEPTION
    WHEN OTHERS THEN
      IF SQLCODE = -955 THEN
        NULL; -- suppresses ORA-00955 exception
      ELSE
         RAISE;
      END IF;
END;
'''
        return query

    def get_drop_table_query(self, schema, table, if_not_exists=None):
        query = f'DROP TABLE {self.mark_to_object(schema)}.{self.mark_to_object(table)}'
        if if_not_exists == 'DoNothing':
            query = f'''declare
v_sql LONG;
begin

v_sql:='{query}';
execute immediate v_sql;

EXCEPTION
    WHEN OTHERS THEN
      IF SQLCODE = -942 THEN
        NULL; -- suppresses ORA-00942 exception
      ELSE
         RAISE;
      END IF;
END;
'''
        return query

    def get_truncate_table_query(self, schema, table):
        return f'TRUNCATE TABLE {self.mark_to_object(schema)}.{self.mark_to_object(table)}'

    def get_schemas(self):
        schemas = self.inspector.get_schema_names()
        return schemas

    def has_table(self, object_name, schema=None):
        result = self.inspector.has_table(table_name=object_name, schema=schema)
        return result

    def get_tables(self, schema):
        tables = self.inspector.get_table_names(schema=schema)
        return tables

    def get_views(self, schema):
        views = self.inspector.get_view_names(schema=schema)
        return views

    def get_columns(self, schema, object_name):
        columns = self.inspector.get_columns(table_name=object_name, schema=schema)
        return columns



from .big_data_connector import BigDataConnector
from .big_data_context import BigDataContext
from .big_data_dialect import BigDataDialect
from .big_data_iterator import BigDataIterator
from .big_data_policy import BigDataPolicy
from .big_data_provider import BigDataProvider



from typing import List

from injector import inject
from pandas import DataFrame

from pdip.integrator.connection.base import ConnectionTargetAdapter
from pdip.integrator.connection.types.bigdata.base import BigDataProvider
from pdip.integrator.integration.domain.base import IntegrationBase


class BigDataTargetAdapter(ConnectionTargetAdapter):
    @inject
    def __init__(self,
                 provider: BigDataProvider,
                 ):
        self.provider = provider

    def clear_data(self, integration: IntegrationBase) -> int:
        target_context = self.provider.get_context_by_config(
            config=integration.TargetConnections.BigData.Connection)
        truncate_affected_rowcount = target_context.truncate_table(schema=integration.TargetConnections.BigData.Schema,
                                                                   table=integration.TargetConnections.BigData.ObjectName)
        return truncate_affected_rowcount

    def prepare_insert_row(self, data, columns):
        insert_rows = []
        for extracted_data in data:
            row = []
            for column in columns:
                column_data = extracted_data[column]
                row.append(column_data)
            insert_rows.append(tuple(row))
        return insert_rows

    def prepare_data(self, integration: IntegrationBase, source_data: any) -> List[any]:
        columns = integration.SourceConnections.Columns
        if columns is not None:
            source_columns = [(column.Name) for column in columns]
        elif columns is None:
            source_columns = source_data[0].keys()
        if isinstance(source_data, DataFrame):
            data = source_data[source_columns]
            prepared_data = data.values.tolist()
        else:
            prepared_data = self.prepare_insert_row(data=source_data, columns=source_columns)
        return prepared_data

    def prepare_target_query(self, integration: IntegrationBase, source_column_count: int) -> str:
        target_context = self.provider.get_context_by_config(
            config=integration.TargetConnections.BigData.Connection)

        columns = integration.SourceConnections.Columns
        if columns is not None and len(columns) > 0:
            source_columns = [(column.Name, column.Type) for column in
                              columns]
            prepared_target_query = target_context.prepare_target_query(
                column_rows=source_columns,
                query=integration.TargetConnections.BigData.Query
            )
        else:
            schema = integration.TargetConnections.BigData.Schema
            table = integration.TargetConnections.BigData.ObjectName
            if schema is None or schema == '' or table is None or table == '':
                raise Exception(f"Schema and table required. {schema}.{table}")
            indexer_array = []
            indexer = target_context.dialect.get_query_indexer()
            for index in range(source_column_count):
                column_indexer = indexer.format(index=index)
                indexer_array.append(column_indexer)
            values_query = ','.join(indexer_array)
            prepared_target_query = target_context.dialect.get_insert_query(values_query=values_query,
                                                                            schema=schema,
                                                                            table=table)
        return prepared_target_query

    def write_data(self, integration: IntegrationBase, prepared_data: List[any]) -> int:
        if prepared_data is not None and len(prepared_data) > 0:
            target_context = self.provider.get_context_by_config(
                config=integration.TargetConnections.BigData.Connection)

            prepared_target_query = self.prepare_target_query(integration=integration,
                                                              source_column_count=len(prepared_data[0]))
            affected_row_count = target_context.execute_many(query=prepared_target_query, data=prepared_data)
            return affected_row_count
        else:
            return 0

    def do_target_operation(self, integration: IntegrationBase) -> int:
        target_context = self.provider.get_context_by_config(
            config=integration.TargetConnections.BigData.Connection)

        affected_rowcount = target_context.execute(query=integration.TargetConnections.BigData.Query)
        return affected_rowcount



from dataclasses import dataclass
from typing import Optional

from ..configuration.base import WebServiceConnectionConfiguration


@dataclass
class ConnectionWebServiceBase:
    Connection: WebServiceConnectionConfiguration = None
    Method: Optional[str] = None
    RequestBody: Optional[str] = None



from ..provider import ServiceProvider


class DependencyContainer:
    Instance: ServiceProvider = None

    @classmethod
    def initialize_service(cls,
                           root_directory=None,
                           configurations: [] = None,
                           excluded_modules: [] = None,
                           initialize_flask: bool = True):
        try:
            cls.Instance = ServiceProvider(root_directory=root_directory,
                                           configurations=configurations,
                                           excluded_modules=excluded_modules)
            cls.Instance.initialize_injection(initialize_flask=initialize_flask)
        except:
            cls.cleanup()
            raise

    @classmethod
    def cleanup(cls):
        if hasattr(cls, 'Instance') and cls.Instance is not None:
            del cls.Instance
            cls.Instance = None



from injector import inject

from pdip.dependency import IScoped
from pdip.logging.loggers.console import ConsoleLogger
from pdip.processing import ProcessManager


class ProcessManagerFactory(IScoped):
    @inject
    def __init__(self, logger: ConsoleLogger):
        self.logger = logger

    def get(self):
        process_manager = ProcessManager(logger=self.logger)
        return process_manager



from .default_operation_execution_initializer import DefaultOperationExecutionInitializer
from .operation_execution_initializer import OperationExecutionInitializer
from .operation_execution_initializer_factory import OperationExecutionInitializerFactory



from dataclasses import dataclass

from .....authentication.basic import ConnectionBasicAuthentication
from .....authentication.kerberos import KerberosAuthentication
from .....authentication.mechanism import MechanismTypes
from .....enums import ConnectionTypes
from .....enums import ConnectorTypes
from .....server.base import ConnectionServer


@dataclass
class BigDataConnectionConfiguration:
    Name: str = None
    ConnectionString: str = None
    ConnectionType: ConnectionTypes = None
    ConnectorType: ConnectorTypes = None
    Driver: str = None
    Server: ConnectionServer = None
    Database: str = None
    BasicAuthentication: ConnectionBasicAuthentication = None
    KerberosAuthentication: KerberosAuthentication = None
    AuthenticationMechanismType: MechanismTypes = None
    Ssl: bool = None
    UseOnlySspi: bool = None
    ApplicationName: str = None



from dataclasses import dataclass
from typing import List, Optional

from ....connection.domain.base import ConnectionColumnBase
from ....connection.domain.enums import ConnectionTypes
from ....connection.domain.types.bigdata.base import ConnectionBigDataBase
from ....connection.domain.types.sql.base import ConnectionSqlBase
from ....connection.domain.types.webservice.base import ConnectionWebServiceBase


@dataclass
class IntegrationConnectionBase:
    ConnectionName: str = None
    ConnectionType: ConnectionTypes = None
    Sql: Optional[ConnectionSqlBase] = None
    BigData: Optional[ConnectionBigDataBase] = None
    WebService: Optional[ConnectionWebServiceBase] = None
    File: Optional[any] = None
    Queue: Optional[any] = None
    Columns: Optional[List[ConnectionColumnBase]] = None


@dataclass
class IntegrationBase:
    SourceConnections: Optional[IntegrationConnectionBase] = None
    TargetConnections: IntegrationConnectionBase = None
    IsTargetTruncate: Optional[bool] = None



from abc import ABC, abstractmethod
from queue import Queue

from pandas import DataFrame

from pdip.dependency import IScoped


class FileConnector(ABC, IScoped):
    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def get_unpredicted_data(self, file: str, names: [], header: int, separator: str, limit: int, process_count: int,
                             data_queue: Queue, result_queue: Queue):
        pass

    @abstractmethod
    def get_data_count(self, file: str):
        pass

    @abstractmethod
    def get_data(self, file: str, names: [], start: int, limit: int,
                 header: int, separator: str) -> DataFrame:
        pass

    @abstractmethod
    def write_data(self, file: str, data: DataFrame, separator: str):
        pass

    @abstractmethod
    def recreate_file(self, file: str, headers: [], separator: str):
        pass

    @abstractmethod
    def delete_file(self, file: str):
        pass



from .sql_connector import SqlConnector


class SqlIterator:

    def __init__(self, connector: SqlConnector, query, excluded_columns=None, limit: int = 0, start: int = 0) -> None:
        self.start = start
        self.offset = 0
        if self.start is not None:
            self.offset = start
        self.excluded_columns = excluded_columns
        self.connector = connector
        self.query = query
        self.cursor = None
        self.limit = limit
        self.excluded_columns = excluded_columns
        self.columns = None
        self.is_finished = False

    def _initialize(self):
        if self.connector.cursor is None:
            self.is_finished = False
            self.connector.connect()
            self.cursor = self.connector.cursor
            self.cursor.execute(self.query)
            self.columns = [column[0] for column in self.cursor.description if
                            self.excluded_columns is None or column[0] not in self.excluded_columns]

    def __iter__(self) -> any:  # this makes the class an Iterable
        return self

    def __next__(self) -> str:
        self._initialize()
        if self.is_finished:
            raise StopIteration

        if self.limit is not None and self.limit > 0:
            data = self.cursor.fetchmany(self.limit)
            if len(data) <= 0:
                self.is_finished = True
                self.connector.disconnect()
                raise StopIteration
            self.offset += len(data)
        else:
            data = self.cursor.fetchall()
            self.is_finished = True
            self.connector.disconnect()
            self.offset += len(data)
        results = [dict(zip(self.columns, t)) for t in data]
        return results



from dataclasses import dataclass


@dataclass
class DataQueueTask:
    Id: int = None
    Data: any = None
    DataTypes: any = None
    IsDataFrame: bool = True
    Start: int = None
    End: int = None
    Limit: int = None
    State: int = None
    IsFinished: bool = None
    IsProcessed: bool = None
    Message: str = None
    Traceback: str = None
    Exception: Exception = None



import traceback
from queue import Queue
from time import time

from pdip.dependency.container import DependencyContainer
from pdip.logging.loggers.console import ConsoleLogger
from pdip.processing.models import ProcessTask


class Subprocess:
    def __init__(
            self,
            initialize_container: bool = None,
            root_directory: str = None
    ):
        if initialize_container:
            DependencyContainer.initialize_service(
                root_directory=root_directory, initialize_flask=False)
            self.logger = DependencyContainer.Instance.get(ConsoleLogger)
        else:
            self.logger = ConsoleLogger()

    def __del__(self) -> None:
        del self.logger

    def start(
            self,
            sub_process_id: int,
            process_queue: Queue,
            process_result_queue: Queue,
            target_method,
            kwargs
    ):
        self.logger.info(f"{sub_process_id} process started")
        try:
            while True:
                process_task: ProcessTask = process_queue.get()
                process_task.SubProcessId = sub_process_id
                process_task.State = 2
                if process_task.IsFinished:
                    self.logger.info(f"{sub_process_id} process finished")
                    # Indicate finished
                    break
                else:
                    # TODO: return result with task

                    start = time()
                    # calling target method
                    kwargs["sub_process_id"] = sub_process_id
                    result = target_method(**kwargs)
                    end = time()
                    self.logger.info(
                        f"{sub_process_id} process finished. time:{end - start}")
                    process_task.IsFinished = True
                    process_task.State = 3
                    process_task.Result = result
                    process_result_queue.put(process_task)
                    return
        except Exception as ex:
            self.logger.error(
                f"{sub_process_id} process getting error:{ex}")
            process_task = ProcessTask(SubProcessId=sub_process_id, State=4, Exception=ex,
                                       Traceback=traceback.format_exc(), IsFinished=True)
            process_result_queue.put(process_task)



import mysql.connector
from sqlalchemy import create_engine
from sqlalchemy.engine import URL

from ...base import SqlConnector
from .....domain.types.sql.configuration.base import SqlConnectionConfiguration


class MysqlConnector(SqlConnector):
    def __init__(self, config: SqlConnectionConfiguration):
        self.config = config
        self.connection = None
        self.cursor = None

    def connect(self):
        self.connection = mysql.connector.connect(user=self.config.BasicAuthentication.User,
                                                  password=self.config.BasicAuthentication.Password,
                                                  database=self.config.Database,
                                                  host=self.config.Server.Host,
                                                  port=self.config.Server.Port)
        self.cursor = self.connection.cursor()

    def disconnect(self):
        try:
            if self.cursor is not None:
                self.cursor.close()

            if self.connection is not None:
                self.connection.close()
        except Exception:
            pass

    def get_connection(self):
        return self.connection

    def get_engine_connection_url(self):
        connection_url = URL.create(
            "mysql",
            username=self.config.BasicAuthentication.User,
            password=self.config.BasicAuthentication.Password,
            host=self.config.Server.Host,
            port=self.config.Server.Port,
            database=self.config.Database,
        )
        return connection_url

    def get_engine(self):
        connection_url = self.get_engine_connection_url()
        engine = create_engine(connection_url)
        return engine

    def execute_many(self, query, data):
        try:
            self.cursor.executemany(query, data)
            self.connection.commit()
            return self.cursor.rowcount
        except Exception as error:
            self.connection.rollback()
            self.cursor.close()
            raise



from dataclasses import dataclass

from .dao_generate_config import DaoGenerateConfig


@dataclass
class GenerateConfig:
    base_directory: str = None
    domain: str = None
    name: str = None
    dao: DaoGenerateConfig = None



import copy

from ...utils.type_checker import TypeChecker


def cls_to_dict(cls):
    def to_dict(self):
        _dict = copy.deepcopy(self.__dict__)
        for attr_name in self.__dict__:
            attr = getattr(self, attr_name)
            if not callable(attr):
                if isinstance(attr, list):
                    entities = []
                    for entity in attr:
                        if (hasattr(entity, 'to_dict')):
                            entities.append(entity.to_dict())
                    _dict[attr_name] = entities
                elif TypeChecker().is_class(attr.__class__) and hasattr(attr, 'to_dict'):
                    _dict[attr_name] = attr.to_dict()
        return _dict

    has_attr = hasattr(cls, 'to_dict')
    if not has_attr:
        setattr(cls, 'to_dict', to_dict)
    return cls



from .base.base_converter import BaseConverter
from .encoders.date_time_encoder import DateTimeEncoder
from .encoders.uuid_encoder import UUIDEncoder
from .encoders.mutliple_json_encoders import MultipleJsonEncoders
from .parsers.date_time_parser import date_time_parser
from .base.json_convert import JsonConvert



from injector import inject

from ..strategies.factories import IntegrationSourceToTargetExecuteStrategyFactory
from ...base import IntegrationAdapter
from ....domain.base import IntegrationBase
from .....connection.factories import ConnectionTargetAdapterFactory
from .....domain.enums.events import EVENT_EXECUTION_INTEGRATION_EXECUTE_TRUNCATE, \
    EVENT_EXECUTION_INTEGRATION_EXECUTE_SOURCE, EVENT_LOG
from .....operation.domain import OperationIntegrationBase
from .....pubsub.base import ChannelQueue
from .....pubsub.domain import TaskMessage
from .....pubsub.publisher import Publisher
from ......dependency import IScoped


class SourceToTargetIntegration(IntegrationAdapter, IScoped):
    @inject
    def __init__(self,
                 integration_execute_strategy_factory: IntegrationSourceToTargetExecuteStrategyFactory,
                 connection_target_adapter_factory: ConnectionTargetAdapterFactory
                 ):
        self.connection_target_adapter_factory = connection_target_adapter_factory
        self.integration_execute_strategy_factory = integration_execute_strategy_factory

    def execute(
            self,
            operation_integration: OperationIntegrationBase,
            channel: ChannelQueue
    ) -> int:
        publisher = Publisher(channel=channel)
        target_adapter = self.connection_target_adapter_factory.get_adapter(
            connection_type=operation_integration.Integration.TargetConnections.ConnectionType
        )
        if operation_integration.Integration.IsTargetTruncate:
            truncate_affected_row_count = target_adapter.clear_data(
                integration=operation_integration.Integration
            )
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_EXECUTION_INTEGRATION_EXECUTE_TRUNCATE,
                    kwargs={
                        "data": operation_integration,
                        "row_count": truncate_affected_row_count
                    }
                )
            )
        order = operation_integration.Order
        process_count = operation_integration.ProcessCount
        limit = operation_integration.Limit
        execute_integration_strategy = self.integration_execute_strategy_factory.get(
            process_count=process_count
        )
        strategy_name = type(execute_integration_strategy).__name__
        publisher.publish(
            message=TaskMessage(
                event=EVENT_LOG,
                kwargs={
                    'data': operation_integration,
                    'message': f"{order} - integration will execute on {strategy_name}. {process_count}-{limit}"
                }
            )
        )
        affected_row_count = execute_integration_strategy.execute(
            operation_integration=operation_integration,
            channel=channel
        )
        publisher.publish(
            message=TaskMessage(
                event=EVENT_EXECUTION_INTEGRATION_EXECUTE_SOURCE,
                kwargs={
                    "data": operation_integration,
                    "row_count": affected_row_count
                }
            )
        )
        return affected_row_count

    def get_start_message(self, integration: IntegrationBase):
        message = f"Integration execute started."
        if integration.TargetConnections.Sql is not None:
            message = f"{integration.TargetConnections.Sql.Schema}.{integration.TargetConnections.Sql.ObjectName} integration execute started."
        elif integration.TargetConnections.BigData is not None:
            message = f"{integration.TargetConnections.BigData.Schema}.{integration.TargetConnections.BigData.ObjectName} integration execute started."
        elif integration.TargetConnections.WebService is not None:
            message = f"{integration.TargetConnections.WebService.Method} integration execute started."
        elif integration.TargetConnections.File is not None:
            message = f"{integration.TargetConnections.File.Folder}\\{integration.TargetConnections.File.FileName} integration execute started."
        elif integration.TargetConnections.Queue is not None:
            message = f"{integration.TargetConnections.Queue.TopicName} integration execute started."
        return message

    def get_finish_message(self, integration: IntegrationBase, data_count: int):
        message = f"Integration execute finished"
        if integration.TargetConnections.Sql is not None:
            message = f"{integration.TargetConnections.Sql.Schema}.{integration.TargetConnections.Sql.ObjectName} integration execute finished."
        elif integration.TargetConnections.BigData is not None:
            message = f"{integration.TargetConnections.BigData.Schema}.{integration.TargetConnections.BigData.ObjectName} integration execute finished."
        elif integration.TargetConnections.BigData is not None:
            message = f"{integration.TargetConnections.WebService.Method} integration execute finished."
        elif integration.TargetConnections.File is not None:
            message = f"{integration.TargetConnections.File.Folder}\\{integration.TargetConnections.File.FileName} integration execute finished."
        elif integration.TargetConnections.Queue is not None:
            message = f"{integration.TargetConnections.Queue.TopicName} integration execute finished."

        return message

    def get_error_message(self, integration: IntegrationBase):
        message = f"Integration execute getting error."
        if integration.TargetConnections.Sql is not None:
            message = f"{integration.TargetConnections.Sql.Schema}.{integration.TargetConnections.Sql.ObjectName} integration execute getting error."
        elif integration.TargetConnections.BigData is not None:
            message = f"{integration.TargetConnections.BigData.Schema}.{integration.TargetConnections.BigData.ObjectName} integration execute getting error."
        if integration.TargetConnections.WebService is not None:
            message = f"{integration.TargetConnections.WebService.Method} integration execute getting error."
        elif integration.TargetConnections.File is not None:
            message = f"{integration.TargetConnections.File.Folder}\\{integration.TargetConnections.File.FileName} integration execute getting error."
        elif integration.TargetConnections.Queue is not None:
            message = f"{integration.TargetConnections.Queue.TopicName} integration execute getting error."
        return message



from cryptography.fernet import Fernet
from injector import inject

from ..configuration.models.application import ApplicationConfig
from ..dependency import IScoped


class CryptoService(IScoped):
    @inject
    def __init__(self, application_config: ApplicationConfig):
        self.application_config = application_config

    def decrypt(self, data):
        encoded_data = data.encode()
        secret_key = self.application_config.secret_key.encode()
        f = Fernet(secret_key)
        decrypted_data = f.decrypt(encoded_data)
        decoded_decrypted_data = decrypted_data.decode()
        return decoded_decrypted_data

    def encrypt(self, data: str):
        encoded_data = data.encode()
        secret_key = self.application_config.secret_key.encode()
        f = Fernet(secret_key)
        encrypted_data = f.encrypt(encoded_data)
        decoded_encrypted_data = encrypted_data.decode()
        return decoded_encrypted_data

    @classmethod
    def generate_key(cls):
        return Fernet.generate_key()



from enum import Enum


class ConnectorTypes(Enum):
    MSSQL = 1
    ORACLE = 2
    POSTGRESQL = 3
    EXCEL = 4
    CSV = 5
    Kafka = 6
    MYSQL = 7
    Impala = 8
    Soap = 9
    CLICKHOUSE = 10
    SqLite = 11



from typing import List

from injector import inject

from ...domain.enums.events import EVENT_EXECUTION_INITIALIZED, EVENT_EXECUTION_STARTED, EVENT_EXECUTION_FINISHED
from ...initializer.execution.operation import OperationExecutionInitializerFactory
from ...integration.base import IntegrationExecution
from ...operation.domain import OperationBase, OperationIntegrationBase
from ...pubsub.base import ChannelQueue
from ...pubsub.domain import TaskMessage
from ...pubsub.publisher import Publisher
from ....data.decorators import transactionhandler
from ....dependency import IScoped


class OperationExecution(IScoped):
    @inject
    def __init__(self,
                 integration_execution: IntegrationExecution,
                 operation_execution_initializer_factory: OperationExecutionInitializerFactory,
                 ):
        self.operation_execution_initializer_factory = operation_execution_initializer_factory
        self.integration_execution = integration_execution

    def __start_execution(self, operation_integrations: List[OperationIntegrationBase], channel: ChannelQueue):
        for operation_integration in operation_integrations:
            self.integration_execution.start(operation_integration=operation_integration, channel=channel)

    @transactionhandler
    def start(self, operation: OperationBase, channel: ChannelQueue):
        publisher = Publisher(channel=channel)
        try:
            initializer = self.operation_execution_initializer_factory.get()
            operation = initializer.initialize(operation)
            publisher.publish(message=TaskMessage(event=EVENT_EXECUTION_INITIALIZED, kwargs={'data': operation}))

            publisher.publish(message=TaskMessage(event=EVENT_EXECUTION_STARTED, kwargs={'data': operation}))
            self.__start_execution(operation_integrations=operation.Integrations, channel=channel)
            publisher.publish(
                message=TaskMessage(event=EVENT_EXECUTION_FINISHED, is_finished=True, kwargs={'data': operation}))
        except Exception as ex:
            publisher.publish(
                message=TaskMessage(event=EVENT_EXECUTION_FINISHED, is_finished=True,
                                    kwargs={'data': operation, 'exception': ex}))
            raise



import cx_Oracle
from sqlalchemy import create_engine

from ...base import SqlConnector
from .....domain.types.sql.configuration.base import SqlConnectionConfiguration


class OracleConnector(SqlConnector):
    def __init__(self, config: SqlConnectionConfiguration):
        self.config = config
        self._tns = None
        self.connection_string = None
        self.connection = None
        self.cursor = None

    def connect(self):
        if self.config.Sid is not None and self.config.Sid != '':
            self._tns = cx_Oracle.makedsn(self.config.Server.Host, self.config.Server.Port,
                                          service_name=self.config.Sid)
        else:
            self._tns = cx_Oracle.makedsn(self.config.Server.Host, self.config.Server.Port,
                                          service_name=self.config.ServiceName)
        self.connection = cx_Oracle.connect(self.config.BasicAuthentication.User,
                                            self.config.BasicAuthentication.Password, self._tns,
                                            encoding="UTF-8",
                                            nencoding="UTF-8")
        self.cursor = self.connection.cursor()

    def disconnect(self):
        try:
            if self.cursor is not None:
                self.cursor.close()

            if self.connection is not None:
                self.connection.close()
        except Exception:
            pass

    def get_connection(self):
        return self.connection

    def get_engine_connection_url(self):
        if self.config.Sid is not None and self.config.Sid != '':
            connection_url = f'oracle+cx_oracle://{self.config.BasicAuthentication.User}:{self.config.BasicAuthentication.Password}@{self.config.Server.Host}:{self.config.Server.Port}/{self.config.Sid}'
        else:
            connection_url = f'oracle+cx_oracle://{self.config.BasicAuthentication.User}:{self.config.BasicAuthentication.Password}@{self.config.Server.Host}:{self.config.Server.Port}/{self.config.ServiceName}'
        return connection_url

    def get_engine(self):
        connection_url = self.get_engine_connection_url()
        engine = create_engine(connection_url)
        return engine

    def execute_many(self, query, data):
        try:
            self.cursor.prepare(query)
            self.cursor.executemany(None, data)
            self.connection.commit()
            return self.cursor.rowcount
        except Exception as error:
            self.connection.rollback()
            self.cursor.close()
            raise



import os
import re
from asyncio import Queue
from os import listdir
from os.path import isfile, join, isdir
from typing import List

from injector import inject
from pandas import DataFrame

from pdip.dependency import IScoped
from ..base import FileConnector


class FileContext(IScoped):
    @inject
    def __init__(self,
                 connector: FileConnector
                 ):
        self.connector: FileConnector = connector

    def get_data_count(self, file):
        count = self.connector.get_data_count(file=file)
        return count

    def get_unpredicted_data(self, file: str, names: [], header: int, separator: str, limit: int, process_count: int,
                             data_queue: Queue, result_queue: Queue):
        data = self.connector.get_unpredicted_data(file=file, names=names, header=header, separator=separator,
                                                   limit=limit, process_count=process_count,
                                                   data_queue=data_queue, result_queue=result_queue)
        return data

    def get_data(self, file: str, names: [], start: int, limit: int, header: int, separator: str) -> DataFrame:

        data = self.connector.get_data(file=file, names=names, start=start, limit=limit, header=header,
                                       separator=separator)

        return data

    def write_to_file(self, file: str, data: DataFrame, separator: str):
        self.connector.write_data(file=file, data=data, separator=separator)

    def recreate_file(self, file: str, headers: [], separator: str):
        self.connector.recreate_file(file=file, headers=headers, separator=separator)

    def delete_file(self, file: str):
        self.connector.delete_file(file=file)

    def get_file_path(self, folder_name: str, file_name: str) -> List[str]:
        file_path = os.path.join(self.connector.host, folder_name, file_name)
        return file_path

    # def get_all_files(self, folder_name: str, file_regex: str) -> List[str]:
    #     folder_path = os.path.join(self.connector.host, folder_name)
    #     regex = re.compile(file_regex)
    #     files = []
    #     for root, dirs, files in os.walk(folder_path):
    #         for file in files:
    #             if regex.match(file):
    #                 files.append(os.path.join(root, file))
    #     return files

    def get_files(self, folder, file_regex) -> List[str]:
        regex = re.compile(file_regex)
        only_files = [f for f in listdir(folder) if isfile(join(folder, f)) and regex.match(f)]
        return only_files

    def get_sub_folders(self, folder):
        only_folders = [f for f in listdir(folder) if isdir(join(folder, f))]
        return only_folders

    def get_all_files(self, folder_name: str, file_regex: str) -> List[str]:
        folder_path = os.path.join(self.connector.host, folder_name)

        sub_folders = self.get_sub_folders(folder_path)
        files = self.get_files(folder_path, file_regex)
        file_list = [join(folder_path, file) for file in files]
        for sub_folder in sub_folders:
            sub_folder_path = join(folder_path, sub_folder)
            files = self.get_files(sub_folder_path, file_regex)
            for file in files:
                file_path = join(sub_folder_path, file)
                file_list.append(file_path)
        return file_list

    def prepare_insert_row(self, data, column_rows):
        insert_rows = []
        for extracted_data in data:
            row = []
            for column_row in column_rows:
                prepared_data = extracted_data[column_rows.index(column_row)]
                row.append(prepared_data)
            insert_rows.append(tuple(row))
        return insert_rows



EVENT_LOG = 0
EVENT_EXECUTION_INITIALIZED = 1
EVENT_EXECUTION_STARTED = 2
EVENT_EXECUTION_FINISHED = 3
EVENT_EXECUTION_INTEGRATION_INITIALIZED = 4
EVENT_EXECUTION_INTEGRATION_STARTED = 5
EVENT_EXECUTION_INTEGRATION_GET_SOURCE_DATA_COUNT = 6
EVENT_EXECUTION_INTEGRATION_EXECUTE_TRUNCATE = 7
EVENT_EXECUTION_INTEGRATION_EXECUTE_TARGET = 8
EVENT_EXECUTION_INTEGRATION_EXECUTE_SOURCE = 9
EVENT_EXECUTION_INTEGRATION_FINISHED = 10



import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from socket import gaierror

from injector import inject

from ..configuration.services import ConfigService
from ..dependency import IScoped
from ..logging.loggers.sql import SqlLogger


class EmailProvider(IScoped):
    @inject
    def __init__(self, config_service: ConfigService, logger: SqlLogger):
        super().__init__()
        self.logger = logger
        self.config_service = config_service

    def send(self, to, subject, body):
        smtp = None
        try:
            host = self.config_service.get_config_by_name("EMAIL_HOST")
            port = self.config_service.get_config_by_name("EMAIL_PORT")
            smtp_address = self.config_service.get_config_by_name("EMAIL_SMTP")
            from_address = self.config_service.get_config_by_name("EMAIL_FROM")
            user = self.config_service.get_config_by_name("EMAIL_USER")
            password = self.config_service.get_config_by_name("EMAIL_PASSWORD")

            if host is None:
                self.logger.error("Email not configured")
                return
            if smtp_address is None:
                self.logger.error("Email smtp not configured")
                return
            if from_address is None:
                self.logger.error("Email from_address not configured")
                return

            # Create the root message and fill in the from, to, and subject headers
            msg_root = MIMEMultipart('related')
            msg_root[
                'Subject'] = subject
            msg_root['From'] = from_address
            recipients = ""
            if isinstance(to, list):
                recipients = ", ".join(to)
            msg_root['To'] = recipients
            msg_root.preamble = 'This is a multi-part message in MIME format.'

            # Encapsulate the plain and HTML versions of the message body in an
            # 'alternative' part, so message agents can decide which they want to display.
            msg_alternative = MIMEMultipart('alternative')
            msg_root.attach(msg_alternative)
            html = body

            # msgText = MIMEText('<b>Some <i>HTML</i> text</b> and an image.<br><img src="cid:image1"><br>Nifty!',
            # 'html') msgAlternative.attach(msgText)
            msg_text = MIMEText(html, "html")
            msg_alternative.attach(msg_text)

            # # This example assumes the image is in the current directory
            # file_path_1 = os.path.join(root_directory, 'image1.jpg')
            # fp_1 = open(file_path_1, 'rb')
            # msgImage_1 = MIMEImage(fp_1.read())
            # fp_1.close()
            # msgImage_1.add_header('Content-ID', '<image1>')
            # msgRoot.attach(msgImage_1)
            try:
                smtp = smtplib.SMTP(host, port)
                # smtp.connect(host, port)
                if user is not None and user != '' and password is not None and password != '':
                    smtp.ehlo()
                    smtp.starttls()
                    smtp.login(user, password)
                smtp.sendmail(smtp_address, to, msg_root.as_string())
            finally:
                if smtp is not None:
                    smtp.quit()

        except (gaierror, ConnectionRefusedError)as ex:
            # tell the script to report if your message was sent or which errors need to be fixed
            self.logger.error('Failed to connect to the server. Bad connection settings? Error:' + str(ex))
        except smtplib.SMTPServerDisconnected as ex:
            self.logger.error('Failed to connect to the server. Wrong user/password? Error:' + str(ex))
        except smtplib.SMTPException as ex:
            self.logger.error('SMTP error occurred: ' + str(ex))
        else:
            self.logger.info('Email sent successfully')
        finally:
            if smtp is not None:
                smtp.close()



from time import time

from injector import inject

from ..types.base import IntegrationAdapter, IntegrationAdapterFactory
from ...domain.enums.events import EVENT_EXECUTION_INTEGRATION_FINISHED, \
    EVENT_EXECUTION_INTEGRATION_STARTED, \
    EVENT_EXECUTION_INTEGRATION_INITIALIZED
from ...initializer.execution.integration import OperationIntegrationExecutionInitializerFactory
from ...operation.domain.operation import OperationIntegrationBase
from ...pubsub.base import ChannelQueue
from ...pubsub.domain import TaskMessage
from ...pubsub.publisher import Publisher
from ....dependency import IScoped


class IntegrationExecution(IScoped):
    @inject
    def __init__(self,
                 integration_adapter_factory: IntegrationAdapterFactory,
                 operation_integration_execution_initializer_factory: OperationIntegrationExecutionInitializerFactory,

                 ):
        self.operation_integration_execution_initializer_factory = operation_integration_execution_initializer_factory
        self.integration_adapter_factory = integration_adapter_factory

    def start(self, operation_integration: OperationIntegrationBase, channel: ChannelQueue):
        start_time = time()
        publisher = Publisher(channel=channel)
        integration_adapter: IntegrationAdapter = self.integration_adapter_factory.get(
            integration=operation_integration.Integration)
        try:
            initializer = self.operation_integration_execution_initializer_factory.get()
            operation_integration = initializer.initialize(operation_integration)
            initialize_message = f'integration initialized.'
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_EXECUTION_INTEGRATION_INITIALIZED,
                    kwargs={
                        'data': operation_integration,
                        'message': initialize_message
                    }
                )
            )

            start_message = integration_adapter.get_start_message(integration=operation_integration.Integration)
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_EXECUTION_INTEGRATION_STARTED,
                    kwargs={
                        'data': operation_integration,
                        'message': start_message
                    }
                )
            )
            data_count = integration_adapter.execute(
                operation_integration=operation_integration,
                channel=channel)

            finish_message = integration_adapter.get_finish_message(integration=operation_integration.Integration,
                                                                    data_count=data_count)

            end_time = time()
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_EXECUTION_INTEGRATION_FINISHED,
                    kwargs={
                        'data': operation_integration,
                        'message': f'{finish_message} time:{end_time - start_time}',
                        'data_count': data_count
                    }
                )
            )
        except Exception as ex:
            error_message = integration_adapter.get_error_message(integration=operation_integration.Integration)
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_EXECUTION_INTEGRATION_FINISHED,
                    kwargs={
                        'data': operation_integration,
                        'message': error_message,
                        'data_count': 0,
                        'exception': ex
                    }
                )
            )
            raise



from flask import request, Response
from injector import inject

from ...configuration.models.api import ApiConfig
from ...data.repository import RepositoryProvider
from ...dependency import ISingleton
from ...dependency.provider import ServiceProvider


class RequestHandler(ISingleton):
    @inject
    def __init__(
            self,
            api_config: ApiConfig,
            service_provider: ServiceProvider
    ):
        self.service_provider = service_provider
        self.api_config = api_config

    def set_headers(self, response):
        white_origin = None
        if self.api_config.origins is not None:
            white_origin = self.api_config.origins.split(',')
        if (white_origin is not None and ((len(white_origin) == 1 and white_origin[0] == '*') or (
                ('Origin' in request.headers and request.headers['Origin'] in white_origin)))):
            response.headers['Access-Control-Allow-Origin'] = request.headers['Origin']
            # response.headers['Access-Control-Allow-Methods'] = 'PUT,GET,POST,DELETE'
            # response.headers['Access-Control-Allow-Headers'] = 'Content-Type,Authorization'
        response.headers['Server'] = ''
        return response

    def after_request(self, response: Response):
        response = self.set_headers(response=response)
        # To terminate database operations at the end of the transaction
        self.service_provider.get(RepositoryProvider).close()
        return response



import os
import re
from multiprocessing.process import current_process


class Utils:

    @staticmethod
    def replace_last(source_string, replace_what, replace_with):
        head, _sep, tail = source_string.rpartition(replace_what)
        return head + replace_with + tail

    @staticmethod
    def to_snake_case(name):
        name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        name = re.sub('__([A-Z])', r'_\1', name)
        name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name)
        return name.lower()

    @staticmethod
    def get_config_name(class_name):
        replace_what = 'Config'
        replace_with = ''
        replaced_name = Utils.replace_last(source_string=class_name, replace_what=replace_what,
                                           replace_with=replace_with)
        snaked_case = Utils.to_snake_case(replaced_name)
        result = snaked_case
        return replaced_name,result

    @staticmethod
    def get_property_name(property_name):
        snaked_case = Utils.to_snake_case(property_name)
        result = snaked_case
        return property_name,result

    @staticmethod
    def path_split(path):
        splits = path.split('/')
        if len(splits) == 1:
            splits = path.split('\\')
        return splits

    @staticmethod
    def get_process_info():
        return f"{current_process().name} ({os.getpid()},{os.getppid()})"

    @staticmethod
    def get_connection_string(database_config, root_directory: str = None):
        if database_config.driver is not None and database_config.driver != '':
            driver = database_config.driver.replace(' ', '+')
        connection_string = ''
        if database_config.type == 'SQLITE':
            connection_type = 'sqlite'
            if database_config.host is not None and database_config.host != '':
                host = database_config.host
                db_path = '/'
                if root_directory is not None:
                    db_path += os.path.join(root_directory, host)
                else:
                    db_path += host
            else:
                db_path = ''
            connection_string = f'{connection_type}://{db_path}'
        if database_config.type == 'MSSQL':
            driver_string = f'?driver={driver}'
            connection_type = 'mssql+pyodbc'
            connection_string = f'{connection_type}://{database_config.user}:{database_config.password}@{database_config.host}:{database_config.port}/{database_config.database}{driver_string}'
        elif database_config.type == 'POSTGRESQL':
            driver_string = ''
            connection_type = 'postgresql'
            connection_string = f'{connection_type}://{database_config.user}:{database_config.password}@{database_config.host}:{database_config.port}/{database_config.database}{driver_string}'
        return connection_string



from queue import Queue
from time import time

import pandas as pd
from func_timeout import FunctionTimedOut, func_set_timeout
from injector import inject
from pandas import DataFrame, notnull

from pdip.dependency import IScoped
from pdip.integrator.connection.domain.task import DataQueueTask
from pdip.integrator.connection.factories import ConnectionSourceAdapterFactory, ConnectionTargetAdapterFactory
from pdip.integrator.domain.enums.events import EVENT_LOG
from pdip.integrator.integration.domain.base import IntegrationBase
from pdip.integrator.operation.domain import OperationIntegrationBase
from pdip.integrator.pubsub.base import ChannelQueue
from pdip.integrator.pubsub.domain import TaskMessage
from pdip.integrator.pubsub.publisher import Publisher


class TargetWriteOperation(IScoped):
    @inject
    def __init__(
            self,
            connection_source_adapter_factory: ConnectionSourceAdapterFactory,
            connection_target_adapter_factory: ConnectionTargetAdapterFactory
    ):
        self.connection_source_adapter_factory = connection_source_adapter_factory
        self.connection_target_adapter_factory = connection_target_adapter_factory

    def start(
            self,
            thread_id: int,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ) -> int:
        publisher = Publisher(channel=channel)
        total_row_count = 0
        try:
            while True:
                data_task: DataQueueTask = data_queue.get()
                if data_task.IsFinished:
                    if data_task.Exception is not None:
                        exc = Exception(data_task.Traceback + '\n' + str(data_task.Exception))
                        raise exc
                    publisher.publish(
                        message=TaskMessage(
                            event=EVENT_LOG,
                            kwargs={
                                'data': operation_integration,
                                'message': f"{thread_id} process tasks finished"
                            }
                        )
                    )
                    return total_row_count
                else:
                    start = time()
                    data = data_task.Data
                    if data_task.IsDataFrame and data is not None:
                        source_data_json = data_task.Data
                        data: DataFrame = DataFrame(source_data_json)
                    data_count = 0
                    if data is None:
                        publisher.publish(
                            message=TaskMessage(
                                event=EVENT_LOG,
                                kwargs={
                                    'data': operation_integration,
                                    'message': f"{thread_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process got a new task without data"
                                }
                            )
                        )
                        data_count = self.start_execute_integration_with_paging(
                            integration=operation_integration.Integration,
                            start=data_task.Start,
                            end=data_task.End
                        )
                    elif data is not None and len(data) > 0:
                        if data_task.IsDataFrame and data_task.DataTypes is not None:
                            source_data = data.astype(dtype=data_task.DataTypes)
                        else:
                            source_data = data
                        if data_task.IsDataFrame:
                            source_data = source_data.where(notnull(data), None)
                            source_data = source_data.replace({pd.NaT: None})

                        publisher.publish(
                            message=TaskMessage(
                                event=EVENT_LOG,
                                kwargs={
                                    'data': operation_integration,
                                    'message': f"{thread_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process got a new task"
                                }
                            )
                        )
                        data_count = self.start_execute_integration_with_source_data(
                            integration=operation_integration.Integration,
                            source_data=source_data
                        )
                    else:
                        publisher.publish(
                            message=TaskMessage(
                                event=EVENT_LOG,
                                kwargs={
                                    'data': operation_integration,
                                    'message': f"{thread_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process got an empty task"
                                }
                            )
                        )
                    total_row_count = total_row_count + data_count
                    end = time()
                    publisher.publish(
                        message=TaskMessage(
                            event=EVENT_LOG,
                            kwargs={
                                'data': operation_integration,
                                'message': f"{thread_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process finished task. time:{end - start}"
                            }
                        )
                    )
                    data_task.IsProcessed = True
                    thread_result = True if data_count > 0 else False
                    data_result_queue.put(thread_result)
                data_queue.task_done()
        except FunctionTimedOut as fto:
            data_result_queue.put(False)
            end = time()
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"{thread_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process getting error. time:{end - start} - Error {fto.getMsg()}"
                    }
                )
            )
            raise Exception(f'Execution Operation timed out after {fto.timedOutAfter} seconds.')
        except Exception as ex:
            data_result_queue.put(False)
            raise

    def start_execute_integration_with_source_data(self,
                                                   integration: IntegrationBase,
                                                   source_data: any
                                                   ):
        target_adapter = self.connection_target_adapter_factory.get_adapter(
            connection_type=integration.TargetConnections.ConnectionType
        )
        self.write_target_data(
            target_adapter=target_adapter,
            integration=integration,
            source_data=source_data
        )
        return len(source_data)

    def start_execute_integration_with_paging(self,
                                              integration: IntegrationBase,
                                              start,
                                              end
                                              ):
        source_adapter = self.connection_source_adapter_factory.get_adapter(
            connection_type=integration.SourceConnections.ConnectionType
        )
        source_data = source_adapter.get_source_data_with_paging(
            integration=integration,
            start=start,
            end=end
        )
        target_adapter = self.connection_target_adapter_factory.get_adapter(
            connection_type=integration.TargetConnections.ConnectionType
        )

        self.write_target_data(
            target_adapter=target_adapter,
            integration=integration,
            source_data=source_data
        )
        return len(source_data)

    @func_set_timeout(1800)
    def write_target_data(
            self,
            target_adapter,
            integration,
            source_data
    ):
        target_adapter.write_data(integration=integration,
                                  source_data=source_data)



import traceback
from datetime import datetime
from logging import CRITICAL, DEBUG, FATAL, ERROR, WARNING, INFO, NOTSET

from injector import inject

from ..base import ILogger
from ..console import ConsoleLogger
from ...domain import LogData
from ....configuration.models.application import ApplicationConfig
from ....configuration.models.database import DatabaseConfig
from ....data.repository import RepositoryProvider
from ....dependency import IScoped
from ....utils.utils import Utils


class SqlLogger(IScoped, ILogger):
    @inject
    def __init__(self,
                 application_config: ApplicationConfig,
                 database_config: DatabaseConfig,
                 console_logger: ConsoleLogger):
        self.console_logger = console_logger
        self.database_config = database_config
        self.application_config = application_config

    def log_to_db(self, level, message, job_id=None):
        logger_class = None
        subclasses = LogData.__subclasses__()
        if subclasses is not None and len(subclasses) > 0:
            logger_class = subclasses[0]

        if logger_class is not None:
            log_datetime = datetime.now()  # datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
            process_info = Utils.get_process_info()
            application_name = ''
            if self.application_config.name is not None:
                application_name = self.application_config.name
            if self.application_config.hostname is not None:
                application_name += f'-{self.application_config.hostname}'
            comment = f'{application_name}-{process_info}'
            try:
                repository_provider = RepositoryProvider(database_config=self.database_config,
                                                         database_session_manager=None)
                log_repository = repository_provider.get(logger_class)
                log = logger_class(TypeId=level, Content=message[0:4000], LogDatetime=log_datetime,
                                   JobId=job_id)
                log_repository.insert(log)
                repository_provider.commit()
            except Exception as ex:
                if job_id is not None:
                    message = f"{job_id}-{message}"
                self.console_logger.error(f'{message}. Sql logging getting error. Error:{ex}')
            finally:
                if job_id is not None:
                    message = f"{job_id}-{message}"
                self.console_logger.log(level, f'{message}')
        else:
            if job_id is not None:
                message = f"{job_id}-{message}"
            self.console_logger.log(level, f'{message}')

    #######################################################################################
    def logger_method(self, level, message, job_id=None):
        self.log_to_db(level=level, message=message, job_id=job_id)

    #######################################################################################
    def exception(self, exception: Exception, message: str = None, job_id=None):
        exc = traceback.format_exc() + '\n' + str(exception)
        if message is not None:
            message += f"Error: {exc}"
        else:
            message = f"Error: {exc}"
        self.logger_method(ERROR, message, job_id)

    #######################################################################################
    def critical(self, message, job_id):
        self.logger_method(CRITICAL, message, job_id)

    #######################################################################################
    def fatal(self, message, job_id=None):
        self.logger_method(FATAL, message, job_id)

    #######################################################################################
    def error(self, message, job_id=None):
        self.logger_method(ERROR, message, job_id)

    #######################################################################################
    def warning(self, message, job_id=None):
        self.logger_method(WARNING, message, job_id)

    #######################################################################################
    def info(self, message, job_id=None):
        self.logger_method(INFO, message, job_id)

    #######################################################################################
    def debug(self, message, job_id=None):
        self.logger_method(DEBUG, message, job_id)

    #######################################################################################
    def log(self, message, job_id=None):
        self.logger_method(NOTSET, message, job_id)



from abc import abstractmethod

from ...base import Initializer
from ....operation.domain import OperationBase


class ExecutionInitializer(Initializer):

    @abstractmethod
    def initialize(
            self,
            operation: OperationBase,
            execution_id: int = None,
            ap_scheduler_job_id: int = None
    ) -> OperationBase:
        pass



import re
import time

from injector import inject

from .in_memory_connector import InMemoryConnector
from .in_memory_policy import InMemoryPolicy
from ......dependency import IScoped


class InMemoryContext(IScoped):
    @inject
    def __init__(self,
                 policy: InMemoryPolicy,
                 retry_count=3):
        self.connector: InMemoryConnector = policy.connector
        self.retry_count = retry_count
        self.default_retry = 1

    def connect(func):
        def inner(*args, **kwargs):
            try:
                args[0].connector.connect()
                return func(*args, **kwargs)
            finally:
                args[0].connector.disconnect()

        return inner

    @connect
    def fetch_query(self, query, excluded_columns=None):
        self.connector.cursor.execute(query)
        if excluded_columns is not None:
            columns = [column[0] for column in self.connector.cursor.description if column[0] not in excluded_columns]
        else:
            columns = [column[0] for column in self.connector.cursor.description]
        results = []
        for row in self.connector.cursor.fetchall():
            results.append(dict(zip(columns, row)))
        # results =  self.connector.cursor.fetchall()
        return results

    @connect
    def execute(self, query) -> any:
        self.connector.cursor.execute(query)
        self.connector.connection.commit()
        return self.connector.cursor.rowcount

    @connect
    def execute_many(self, query, data):
        return self._execute_with_retry(query=query, data=data, retry=self.default_retry)

    def _execute_many_start(self, query, data):
        return self.connector.execute_many(query=query, data=data)

    def _execute_with_retry(self, query, data, retry):
        try:
            return self._execute_many_start(query=query, data=data)
        except Exception as ex:
            if retry > self.retry_count:
                print(f"Db write error on Error:{ex}")
                raise
            print(
                f"Getting error on insert (Operation will be retried. Retry Count:{retry}). Error:{ex}")
            # retrying connect to db,
            self.connector.connect()
            time.sleep(1)
            return self._execute_with_retry(query=query, data=data, retry=retry + 1)

    @connect
    def get_table_count(self, query):
        count_query = self.connector.get_table_count_query(query=query)
        self.connector.cursor.execute(count_query)
        datas = self.connector.cursor.fetchall()
        return datas[0][0]

    def get_table_data(self, query):
        data_query = self.connector.get_table_data_query(query=query)
        return self.fetch_query(data_query)

    def get_table_data_with_paging(self, query, start, end):
        data_query = self.connector.get_table_data_with_paging_query(query=query,
                                                                     start=start,
                                                                     end=end)
        results = self.fetch_query(data_query, excluded_columns=['row_number'])

        return results

    def truncate_table(self, schema, table):
        truncate_query = self.connector.get_truncate_query(schema=schema, table=table)
        return self.execute(query=truncate_query)

    @staticmethod
    def replace_regex(text, field, indexer):
        text = re.sub(r'\(:' + field + r'\b', f'({indexer}', text)
        text = re.sub(r':' + field + r'\b\)', f'{indexer})', text)
        text = re.sub(r':' + field + r'\b', f'{indexer}', text)
        return text

    def prepare_target_query(self, column_rows, query):
        target_query = query
        for column_row in column_rows:
            index = column_rows.index(column_row)
            indexer = self.connector.get_target_query_indexer().format(index=index)
            target_query = self.replace_regex(target_query, column_row[0], indexer)
        return target_query

    def prepare_insert_row(self, data, column_rows):
        insert_rows = []
        for extracted_data in data:
            row = []
            for column_row in column_rows:
                prepared_data = extracted_data[column_rows.index(column_row)]
                row.append(prepared_data)
            insert_rows.append(tuple(row))
        return insert_rows



from injector import inject

from .default_integrator_initializer import DefaultIntegratorInitializer
from .integrator_initializer import IntegratorInitializer
from ....dependency import IScoped
from ....dependency.provider import ServiceProvider


class IntegratorInitializerFactory(IScoped):
    @inject
    def __init__(
            self,
            service_provider: ServiceProvider
    ):
        self.service_provider = service_provider

    def get(self) -> IntegratorInitializer:
        subclasses = IntegratorInitializer.__subclasses__()
        if subclasses is not None and len(subclasses) > 0:
            if len(subclasses) > 1:
                initializer_classes = [subclass for subclass in subclasses if subclass != DefaultIntegratorInitializer]
                initializer_class = initializer_classes[0]
            else:
                initializer_class = subclasses[0]
            initializer = self.service_provider.get(initializer_class)
            return initializer



from abc import abstractmethod


class SqlDialect:
    def __init__(self):
        pass

    @property
    @abstractmethod
    def indexer(self) -> str:
        pass

    @property
    @abstractmethod
    def quotation_mark(self) -> str:
        pass

    @abstractmethod
    def mark_to_object(self, o) -> str:
        pass

    @abstractmethod
    def get_table_count_query(self, schema, table):
        pass

    @abstractmethod
    def get_count_query(self, query):
        pass

    @abstractmethod
    def get_select_query(self, query):
        pass

    @abstractmethod
    def get_table_select_query(self, schema, table, selected_rows):
        pass

    @abstractmethod
    def get_paging_query(self, query, start, end):
        pass

    @abstractmethod
    def prepare_select_query(self, schema, table, columns=None):
        pass

    @abstractmethod
    def generate_insert_query(self, schema, table, columns_query, values_query):
        pass

    @abstractmethod
    def get_insert_query(self, schema, table, columns_query, values_query):
        pass

    @abstractmethod
    def get_insert_values_query(self, schema, table, values_query):
        pass

    @abstractmethod
    def get_create_table_query(self, schema, table, columns, if_exists=None):
        pass

    @abstractmethod
    def get_drop_table_query(self, schema, table, if_not_exists=None):
        pass

    @abstractmethod
    def get_truncate_table_query(self, schema, table):
        pass

    @abstractmethod
    def get_schemas(self):
        pass

    @abstractmethod
    def has_table(self, object_name, schema=None):
        pass

    @abstractmethod
    def get_tables(self, schema):
        pass

    @abstractmethod
    def get_sorted_table_and_fkc_names(self, schema):
        pass

    @abstractmethod
    def get_views(self, schema):
        pass

    @abstractmethod
    def get_columns(self, schema, table):
        pass




from abc import abstractmethod

from ...base import Initializer
from ....operation.domain import OperationIntegrationBase


class OperationIntegrationExecutionInitializer(Initializer):
    @abstractmethod
    def initialize(self, operation_integration: OperationIntegrationBase) -> OperationIntegrationBase:
        pass



import json
import uuid


class UUIDEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            # if the obj is uuid, we simply return the value of uuid
            return obj.hex
        return super().default(obj)



import queue


class ChannelQueue:
    def __init__(self, channel_queue: queue):
        self.channel_queue: queue = channel_queue

    def put(self, message):
        self.channel_queue.put(message)

    def get(self):
        return self.channel_queue.get()

    def done(self):
        return self.channel_queue.task_done()



import glob
import os
from typing import List, Type, TypeVar

import yaml

from .models.base.base_config import BaseConfig
from ..utils import Utils, ModuleFinder

T = TypeVar('T')


class ConfigManager:
    def __init__(self, root_directory: str, module_finder: ModuleFinder) -> None:
        # Create an empty list with items of type T
        self.module_finder = module_finder
        self.configs = self.__get_configs(root_directory)

    def get_by_name(self, name):
        for config in self.configs:
            if config["name"] == name:
                return config.get("instance")

    def get_all(self):
        return self.configs

    def get_all_type_configs(self):
        return [c for c in self.configs if c["type"] is not None]

    def get(self, generic_type: Type[T]) -> T:
        for config in self.configs:
            config_type = config.get("type")
            if config_type is not None and (isinstance(config.get("instance"),
                          generic_type) or config_type is generic_type or generic_type.__module__ == config_type.__module__):
                return config.get("instance")

    def set(self, generic_type, instance_property, property_value):
        config_instance = self.get(generic_type=generic_type)
        if config_instance is None:
            config_instance = generic_type()
            self.configs.append({"type": generic_type, "instance": config_instance})
        setattr(config_instance, instance_property, property_value)

    def __get_configs(self, root_directory: str) -> List[dict]:
        self.module_finder.import_modules_by_name_ends_with(name='Configs')

        environment = os.getenv('PYTHON_ENVIRONMENT', None)
        config_name = "application.yml"
        if environment is not None and environment != '':
            config_name_split = "application.yml".split('.')
            config_name = f'{config_name_split[0]}.{environment}.{config_name_split[1]}'
        config_path = os.path.join(root_directory, config_name)
        configs: List[dict] = []
        if not os.path.exists(config_path):
            config_files = glob.glob(os.path.join(root_directory, "application.*.yml"))
            if len(config_files) > 0:
                config_path = config_files[0]
        if os.path.exists(config_path):
            with open(config_path, 'r') as yml_file:
                loaded_configs = yaml.load(yml_file, Loader=yaml.FullLoader)
            for config in BaseConfig.__subclasses__():
                config_instance = config()
                class_name, class_sneak_name = Utils.get_config_name(config_instance.__class__.__name__)
                search_class_in_configs = [class_name, class_sneak_name, class_name.upper(), class_sneak_name.upper()]
                class_properties = [a for a in dir(
                    config_instance) if not (a.startswith('_'))]
                loaded_config=None
                for search_in_config in search_class_in_configs:
                    if search_in_config in loaded_configs:
                        loaded_config = loaded_configs[search_in_config]
                        break
                if loaded_config is None:
                    continue
                for prop in class_properties:
                    config_value = None
                    property_name, property_snake_name = Utils.get_property_name(prop)
                    search_property_in_configs = [property_name, property_snake_name, property_name.upper(),
                                                  property_snake_name.upper()]

                    for search_property_in_config in search_property_in_configs:
                        if search_property_in_config in loaded_config:
                            config_value = loaded_config[search_property_in_config]
                            break
                    if property_snake_name.upper() == 'ROOT_DIRECTORY':
                        config_value = root_directory

                    environment_name = f'{class_sneak_name.upper()}_{property_snake_name.upper()}'
                    property_value = os.getenv(environment_name, config_value)
                    setattr(config_instance, prop, property_value)
                configs.append({"name": class_name, "type": config, "instance": config_instance})
            for key in loaded_configs.keys():
                has_key = False
                class_name, class_sneak_name = Utils.get_config_name(key)
                search_class_in_configs = [class_name, class_sneak_name, class_name.upper(), class_sneak_name.upper()]

                for conf in configs:
                    config_name, config_sneak_name = Utils.get_config_name(conf["name"])
                    config_names = [config_name, config_sneak_name, config_name.upper(), config_sneak_name.upper()]
                    for search_class_in_config in search_class_in_configs:
                        if search_class_in_config in config_names:
                            has_key = True
                            break

                if not has_key:
                    configs.append({"name": key, "type": None, "instance": loaded_configs[key]})
        return configs



from abc import abstractmethod
from typing import List

from ...integration.domain.base import IntegrationBase


class ConnectionSourceAdapter:

    @abstractmethod
    def get_source_data_count(
            self,
            integration: IntegrationBase
    ) -> int:
        pass

    @abstractmethod
    def get_iterator(
            self,
            integration: IntegrationBase,
            limit: int
    ) -> List[any]:
        pass

    @abstractmethod
    def get_source_data_with_paging(
            self,
            integration: IntegrationBase,
            start: int,
            end: int
    ) -> List[any]:
        pass



import csv
import json
import os
from queue import Queue

import pandas as pd
from pandas import DataFrame

from pdip.integrator.connection.domain import DataQueueTask
from ...base import FileConnector


class CsvConnector(FileConnector):
    def __init__(self,
                 host: str):
        self.host = host
        self.data_frame = None

    def connect(self):
        # os.path.isdir(self.folder)
        pass

    def disconnect(self):
        pass
        # try:
        #     if self.file is not None:
        #         self.file.close()
        # except Exception:
        #     pass

    def get_unpredicted_data(self, file: str, names: [], header: int, separator: str, limit: int, process_count: int,
                             data_queue: Queue,
                             result_queue: Queue):
        file_path = file
        total_data_count = 0
        transmitted_data_count = 0
        task_id = 0
        for chunk in pd.read_csv(file_path, names=names, decimal=',', sep=separator, header=header, chunksize=limit,
                                 iterator=True, low_memory=False):
            data = json.loads(chunk.to_json(orient='records', date_format="iso"))
            task_id = task_id + 1
            data_count = len(chunk)
            total_data_count = total_data_count + data_count
            data_queue_task = DataQueueTask(Id=task_id, Data=data, Start=total_data_count - data_count,
                                            End=total_data_count, Limit=limit, Message=f'{file_path} file readed',
                                            IsFinished=False)
            data_queue.put(data_queue_task)
            transmitted_data_count = transmitted_data_count + 1
            if transmitted_data_count >= process_count:
                result = result_queue.get()
                if result:
                    transmitted_data_count = transmitted_data_count - 1
                else:
                    break

    def get_data_count(self, file: str):
        file_path = os.path.join(self.host, file)
        with open(file_path, 'rb') as f:
            count = sum(1 for line in f)
        return count

    def get_data(self, file: str, names: [], start: int, limit: int, header: int,
                 separator: str) -> DataFrame:
        file_path = os.path.join(self.host, file)
        count = 0
        for chunk in pd.read_csv(file_path, names=names, sep=separator, header=header, chunksize=limit, iterator=True,
                                 low_memory=False):
            count += len(chunk)
            if start < count:
                return chunk

        return None

    def write_data(self, file: str, data: DataFrame, separator: str):
        file_path = os.path.join(self.host, file)
        data.to_csv(file_path, mode='a', header=0, sep=separator, index=False)

    def recreate_file(self, file: str, headers: [], separator: str):
        self.delete_file(file=file)
        file_path = os.path.join(self.host, file)
        with open(file_path, 'w', newline='') as f:
            writer = csv.writer(f, delimiter=separator)
            writer.writerow(headers)

    def delete_file(self, file: str):
        file_path = os.path.join(self.host, file)
        if os.path.exists(file_path):
            os.remove(file_path)



from pdip.data.domain import EntityBase


class ConfigParameterBase(EntityBase):

    def __init__(self,
                 Name: str = None,
                 Type: str = None,
                 Value: str = None,
                 Description: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.Name: str = Name
        self.Type: str = Type
        self.Value: str = Value
        self.Description: str = Description



from typing import List

from injector import inject

from pdip.exceptions import NotSupportedFeatureException
from pdip.integrator.connection.base import ConnectionSourceAdapter
from pdip.integrator.connection.types.webservice.base import WebServiceProvider
from pdip.integrator.integration.domain.base import IntegrationBase


class WebServiceSourceAdapter(ConnectionSourceAdapter):
    @inject
    def __init__(self,
                 provider: WebServiceProvider,
                 ):
        self.provider = provider

    def get_source_data_count(self, integration: IntegrationBase) -> int:
        raise NotSupportedFeatureException(f"{self.__class__.__name__} get_source_data_count")

    def get_source_data(self, integration: IntegrationBase) -> List[any]:
        raise NotSupportedFeatureException(f"{self.__class__.__name__} get_source_data")

    def get_source_data_with_paging(self, integration: IntegrationBase, start, end) -> List[any]:
        raise NotSupportedFeatureException(f"{self.__class__.__name__} get_source_data_with_paging")



def controller(namespace=None, namespace_name=None, route=None):
    def decorate(cls):
        cls.namespace = namespace
        cls.namespace_name = namespace_name
        cls.route = route
        return cls

    return decorate



from injector import inject

from .sql_context import SqlContext
from .sql_policy import SqlPolicy
from ....domain.authentication.basic import ConnectionBasicAuthentication
from ....domain.enums import ConnectorTypes, ConnectionTypes
from ....domain.server.base import ConnectionServer
from ....domain.types.sql.configuration.base import SqlConnectionConfiguration
from ......dependency import IScoped


class SqlProvider(IScoped):
    @inject
    def __init__(self):
        pass

    def __initialize_context(self, config: SqlConnectionConfiguration):
        policy = SqlPolicy(config=config)
        context: SqlContext = SqlContext(policy=policy)
        return context

    def get_context_by_config(self, config: SqlConnectionConfiguration) -> SqlContext:
        return self.__initialize_context(config=config)

    def get_context(self, connector_type: ConnectorTypes, host: str, port: int, user: str, password: str,
                    database: str = None, service_name: str = None, sid: str = None) -> SqlContext:
        """
        Creating Context
        """
        if connector_type == ConnectorTypes.ORACLE:
            config = SqlConnectionConfiguration(ConnectionType=ConnectionTypes.Sql,
                                                ConnectorType=connector_type.ORACLE,
                                                Server=ConnectionServer(Host=host, Port=port),
                                                BasicAuthentication=ConnectionBasicAuthentication(User=user,
                                                                                                  Password=password),
                                                Sid=sid, ServiceName=service_name)
        elif connector_type == ConnectorTypes.MSSQL:
            config = SqlConnectionConfiguration(ConnectionType=ConnectionTypes.Sql,
                                                ConnectorType=ConnectorTypes.MSSQL,
                                                Server=ConnectionServer(Host=host, Port=port),
                                                BasicAuthentication=ConnectionBasicAuthentication(User=user,
                                                                                                  Password=password),
                                                Database=database)
        elif connector_type == ConnectorTypes.POSTGRESQL:
            config = SqlConnectionConfiguration(ConnectionType=ConnectionTypes.Sql,
                                                ConnectorType=ConnectorTypes.POSTGRESQL,
                                                Server=ConnectionServer(Host=host, Port=port),
                                                BasicAuthentication=ConnectionBasicAuthentication(User=user,
                                                                                                  Password=password),
                                                Database=database)
        elif connector_type == ConnectorTypes.MYSQL:
            config = SqlConnectionConfiguration(ConnectionType=ConnectionTypes.Sql,
                                                ConnectorType=ConnectorTypes.MYSQL,
                                                Server=ConnectionServer(Host=host, Port=port),
                                                BasicAuthentication=ConnectionBasicAuthentication(User=user,
                                                                                                  Password=password),
                                                Database=database)
        elif connector_type == ConnectorTypes.CLICKHOUSE:
            config = SqlConnectionConfiguration(ConnectionType=ConnectionTypes.Sql,
                                                ConnectorType=ConnectorTypes.CLICKHOUSE,
                                                Server=ConnectionServer(Host=host, Port=port),
                                                BasicAuthentication=ConnectionBasicAuthentication(User=user,
                                                                                                  Password=password),
                                                Database=database)
        else:
            raise Exception(f"{connector_type.name} connector type not supported")

        return self.__initialize_context(config=config)



from injector import inject

from ..base import ConnectionSourceAdapter
from ..domain.enums import ConnectionTypes
from ..types.bigdata.adapters.source import BigDataSourceAdapter
from ..types.sql.adapters.source import SqlSourceAdapter
from ..types.webservice.adapters.source import WebServiceSourceAdapter
from ....dependency import IScoped
from ....exceptions import IncompatibleAdapterException, NotSupportedFeatureException


class ConnectionSourceAdapterFactory(IScoped):
    @inject
    def __init__(self,
                 sql_source_adapter: SqlSourceAdapter,
                 big_data_source_adapter: BigDataSourceAdapter,
                 web_service_source_adapter: WebServiceSourceAdapter,
                 # file_adapter: FileAdapter,
                 # queue_adapter: QueueAdapter,
                 ):
        # self.queue_adapter = queue_adapter
        # self.file_adapter = file_adapter
        self.web_service_source_adapter = web_service_source_adapter
        self.big_data_source_adapter = big_data_source_adapter
        self.sql_source_adapter = sql_source_adapter

    def get_adapter(self, connection_type: ConnectionTypes) -> ConnectionSourceAdapter:
        if connection_type == ConnectionTypes.Sql:
            if isinstance(self.sql_source_adapter, ConnectionSourceAdapter):
                return self.sql_source_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.sql_source_adapter} is incompatible with ConnectionSourceAdapter")
        elif connection_type == ConnectionTypes.File:
            if isinstance(self.file_source_adapter, ConnectionSourceAdapter):
                return self.file_source_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.file_source_adapter} is incompatible with ConnectionSourceAdapter")
        elif connection_type == ConnectionTypes.Queue:
            if isinstance(self.queue_source_adapter, ConnectionSourceAdapter):
                return self.queue_source_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.queue_source_adapter} is incompatible with ConnectionSourceAdapter")
        elif connection_type == ConnectionTypes.BigData:
            if isinstance(self.big_data_source_adapter, ConnectionSourceAdapter):
                return self.big_data_source_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.big_data_source_adapter} is incompatible with ConnectionSourceAdapter")
        elif connection_type == ConnectionTypes.WebService:
            if isinstance(self.web_service_source_adapter, ConnectionSourceAdapter):
                return self.web_service_source_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.web_service_source_adapter} is incompatible with ConnectionSourceAdapter")
        else:
            raise NotSupportedFeatureException(f"{connection_type.name}")



import re
from datetime import datetime


def date_time_parser(dct):
    for k, v in dct.items():
        reg_T = re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.{0,1}\d{0,6}$')
        reg_iso = re.compile(r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.{0,1}\d{0,6}$')
        if isinstance(v, str) and reg_T.match(v):
            dct[k] = datetime.strptime(v, "%Y-%m-%dT%H:%M:%S.%f")
        elif isinstance(v, str) and reg_iso.match(v):
            dct[k] = datetime.strptime(v, "%Y-%m-%d %H:%M:%S.%f")
    return dct



from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import dataclass_json

from ...execution.domain import ExecutionOperationIntegrationBase, ExecutionOperationBase
from ...integration.domain.base import IntegrationBase


@dataclass_json
@dataclass
class OperationIntegrationBase:
    Id: Optional[int] = None
    Name: str = None
    Order: int = None
    Limit: Optional[int] = None
    ProcessCount: Optional[int] = None
    Integration: IntegrationBase = None
    Execution: Optional[ExecutionOperationIntegrationBase] = None


@dataclass_json
@dataclass
class OperationBase:
    Id: Optional[int] = None
    Name: str = None
    Integrations: List[OperationIntegrationBase] = None
    Execution: Optional[ExecutionOperationBase] = None



from injector import inject

from pdip.dependency import IScoped
from pdip.integrator.connection.domain.enums import ConnectorTypes
from .queue_connector import QueueConnector
from .queue_context import QueueContext
from ..connectors import KafkaConnector


class QueueProvider(IScoped):
    @inject
    def __init__(self,
                 ):
        pass

    def get_context(self, connection, connection_basic_authentication, connection_servers) -> QueueContext:
        """
        Creating Connection
        """
        connector: QueueConnector = None
        if connection.Queue.ConnectorType.Name == ConnectorTypes.Kafka.name:
            servers = []
            for connection_server in connection_servers:
                server = f"{connection_server.Host}:{connection_server.Port}"
                servers.append(server)
            auth = None
            if ((connection.Queue.Protocol is not None and connection.Queue.Protocol != '') and (
                    connection.Queue.Mechanism is not None and connection.Queue.Mechanism != '') and (
                    connection_basic_authentication.User is not None and connection_basic_authentication.User != '') and (
                    connection_basic_authentication.Password is not None and connection_basic_authentication.Password != '')):
                auth = {
                    'security_protocol': connection.Queue.Protocol,
                    'sasl_mechanism': connection.Queue.Mechanism,
                    'sasl_plain_username': connection_basic_authentication.User,
                    'sasl_plain_password': connection_basic_authentication.Password
                }
            connector = KafkaConnector(servers=servers, auth=auth)
        if connector is not None:
            queue_context: QueueContext = QueueContext(connector=connector)
            return queue_context
        else:
            raise Exception(f"{connection.Queue.ConnectorType.Name} connector type not supported")



from abc import abstractmethod


class SqlConnector(object):
    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def get_engine_connection_url(self):
        pass

    @abstractmethod
    def get_engine(self):
        pass

    @abstractmethod
    def get_connection(self):
        pass

    @abstractmethod
    def execute_many(self, query, data):
        pass



import os
import sqlite3

from pdip.integrator.connection.domain.types.inmemory import InMemoryConnectionConfiguration
from ...base import InMemoryConnector


class SqLiteConnector(InMemoryConnector):
    def __init__(self, config: InMemoryConnectionConfiguration):
        self.config = config
        self.connection_string = None
        if self.config.ConnectionString is not None:
            self.connection_string = self.config.ConnectionString
        else:
            database_directory = os.getcwd()
            self.connection_string = os.path.join(database_directory, self.config.Database)
        self.connection = None
        self.cursor = None

    def connect(self):
        self.connection = sqlite3.connect(self.connection_string)
        self.cursor = self.connection.cursor()

    def disconnect(self):
        try:
            if self.cursor is not None:
                self.cursor.close()
            if self.connection is not None:
                self.connection.close()
        except Exception:
            pass

    def get_connection(self):
        return self.connection

    def execute_many(self, query, data):
        try:
            self.cursor.executemany(query, data)
            self.connection.commit()
            return self.cursor.rowcount
        except Exception as error:
            self.connection.rollback()
            self.cursor.close()

    def get_target_query_indexer(self):
        indexer = '?'
        return indexer



from typing import Type, TypeVar

from injector import inject
from sqlalchemy.orm import Query

from .repository import Repository
from ..base import DatabaseSessionManager
from ...configuration.models.database import DatabaseConfig
from ...dependency import IScoped

R = TypeVar('R')


class RepositoryProvider(IScoped):
    @inject
    def __init__(self,
                 database_config: DatabaseConfig,
                 database_session_manager: DatabaseSessionManager
                 ):
        self.database_config = database_config
        self.database_session_manager = database_session_manager

    # def __del__(self):
    #     self.close()

    def create(self) -> DatabaseSessionManager:
        if self.database_session_manager is None:
            self.database_session_manager = DatabaseSessionManager(database_config=self.database_config)
            return self.database_session_manager
        else:
            return self.database_session_manager

    def get(self, repository_type: Type[R]) -> Repository[R]:
        self.create()
        repository = Repository[repository_type](repository_type, self.database_session_manager)
        return repository

    def query(self, *entities, **kwargs) -> Query:
        database_session_manager = self.create()
        return database_session_manager.session.query(*entities, **kwargs)

    def commit(self):
        if self.database_session_manager is not None:
            self.database_session_manager.commit()

    def rollback(self):
        if self.database_session_manager is not None:
            self.database_session_manager.rollback()

    def reconnect(self):
        if self.database_session_manager is not None:
            self.database_session_manager.close()
            self.database_session_manager.connect()

    def close(self):
        if self.database_session_manager is not None:
            self.database_session_manager.close()
            # self.database_session_manager = None



from datetime import datetime

from ...data.domain import EntityBase


class LogData(EntityBase):
    def __init__(self,
                 TypeId: int = None,
                 Content: str = None,
                 LogDatetime: datetime = None,
                 JobId: int = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.TypeId = TypeId
        self.Content = Content
        self.LogDatetime = LogDatetime
        self.JobId = JobId



import time

from injector import inject

from .integrator_event_manager import IntegratorEventManager
from ...operation.domain import OperationBase, OperationIntegrationBase
from ....dependency import IScoped
from ....logging.loggers.console import ConsoleLogger


class DefaultIntegratorEventManager(IScoped, IntegratorEventManager):
    @inject
    def __init__(self,
                 logger: ConsoleLogger):
        self.logger = logger

    def log(self, data: any, message: str, exception=None):
        if exception is not None:
            if isinstance(data, OperationBase):
                log_message = f'{data.Name} - {message}'
            else:
                log_message = f'{data.Order} - {data.Name} - {message}'

            self.logger.exception(exception, log_message)
        else:
            if isinstance(data, OperationBase):
                log_message = f'{data.Name} - {message}'
            else:
                log_message = f'{data.Order} - {data.Name} - {message}'
            self.logger.info(log_message)

    def initialize(self, data: OperationBase):
        message = f'{data.Name} initialized.'
        self.logger.info(message)

    def start(self, data: OperationBase):
        message = f'{data.Name} started.'
        self.logger.info(message)

    def finish(self, data: OperationBase, exception=None):
        if exception is not None:
            message = f'{data.Name} finished with error.'
            self.logger.exception(exception, message)
        else:
            message = f'{data.Name} finished.'
            self.logger.info(message)

    def integration_initialize(self, data: OperationIntegrationBase, message):
        message = f'{data.Order} - {data.Name} - {message}'
        self.logger.info(message)

    def integration_start(self, data: OperationIntegrationBase, message):
        message = f'{data.Order} - {data.Name} - {message}'
        self.logger.info(message)

    def integration_finish(self, data: OperationIntegrationBase, data_count, message, exception=None):
        if exception is not None:
            message = f'{data.Order} - {data.Name} - {message}'
            self.logger.exception(exception, message)
        else:
            message = f'{data.Order} - {data.Name} - {message}'
            self.logger.info(message)

    def integration_target_truncate(self, data: OperationIntegrationBase, row_count):
        message = f'{data.Order} - {data.Name} - Target truncate finished. (Affected Row Count:{row_count})'
        self.logger.info(message)

    def integration_execute_source(self, data: OperationIntegrationBase, row_count):
        message = f'{data.Order} - {data.Name} - Source integration completed. (Source Data Count:{row_count})'
        self.logger.info(message)

    def integration_execute_target(self, data: OperationIntegrationBase, row_count):
        message = f'{data.Order} - {data.Name} - Target integration completed. (Affected Row Count:{row_count})'
        self.logger.info(message)
        time.sleep(2)



from dataclasses import dataclass
from datetime import datetime
from typing import List

from ...domain.enums import StatusTypes


@dataclass
class ExecutionBase:
    Id: int = None
    ApSchedulerJobId: int = None
    Status: StatusTypes = None
    StartDate: datetime = None
    EndDate: datetime = None


@dataclass
class EventBase:
    Id: int = None
    EventId: int = None
    EventDate: datetime = None


@dataclass
class ExecutionOperationIntegrationEvent(EventBase):
    pass


@dataclass
class ExecutionOperationIntegrationBase(ExecutionBase):
    Name: str = None
    OperationId: int = None
    OperationExecutionId: int = None
    OperationIntegrationId: int = None
    Events: List[ExecutionOperationIntegrationEvent] = None


@dataclass
class ExecutionOperationEvent(EventBase):
    OperationId: int = None
    Status: StatusTypes = None
    Event: int = None


@dataclass
class ExecutionOperationBase(ExecutionBase):
    Name: str = None
    OperationId: int = None
    Events: List[ExecutionOperationEvent] = None



from typing import Generic, TypeVar, Union

from .icommand import ICommand
from .iquery import IQuery

CQH = TypeVar('CQH', covariant=True, bound=Union[IQuery, ICommand])


class CommandQueryHandlerBase(Generic[CQH]):
    def __init__(self) -> CQH:
        pass

    def handle(self, query: CQH) -> any:
        pass



from queue import Queue

from injector import inject
from pandas import DataFrame

from pdip.dependency import IScoped
from .queue_connector import QueueConnector


class QueueContext(IScoped):
    @inject
    def __init__(self,
                 connector: QueueConnector
                 ):
        self.connector: QueueConnector = connector

    def create_topic(self, topic_name):
        return self.connector.create_topic(topic_name=topic_name)

    def delete_topic(self, topic_name):
        return self.connector.delete_topic(topic_name=topic_name)

    def write_data(self, topic_name: str, messages: DataFrame):
        return self.connector.write_data(topic_name=topic_name, messages=messages)

    def get_unpredicted_data(self, topic_name: str, group_id: str, limit: int, process_count: int, data_queue: Queue,
                             result_queue: Queue,
                             auto_offset_reset: str = 'earliest',
                             enable_auto_commit: bool = True):
        self.connector.create_consumer(topic_name=topic_name, group_id=group_id, auto_offset_reset=auto_offset_reset,
                                       enable_auto_commit=enable_auto_commit)
        data = self.connector.get_unpredicted_data(limit=limit, process_count=process_count, data_queue=data_queue,
                                                   result_queue=result_queue)
        return data

    def get_data(self, topic_name: str, group_id: str, limit: int, auto_offset_reset: str = 'earliest',
                 enable_auto_commit: bool = True):
        self.connector.create_consumer(topic_name=topic_name, group_id=group_id, auto_offset_reset=auto_offset_reset,
                                       enable_auto_commit=enable_auto_commit)
        data = self.connector.get_data(limit=limit)
        return data

    def prepare_insert_row(self, data, column_rows):
        insert_rows = []
        for extracted_data in data:
            row = []
            for column_row in column_rows:
                prepared_data = extracted_data[column_rows.index(column_row)]
                row.append(prepared_data)
            insert_rows.append(tuple(row))
        return insert_rows
